def _move_col_after(df, col, col_ref):
    if col in df.columns:
        if col_ref in df.columns:
            loc_var = df.columns.get_loc(col_ref)
            col2insert = df[col]
            df = df.drop([col], axis=1)
            df.insert(loc_var + 1, col, col2insert)
    return df


def _move_col_before(df, col, col_ref):
    if col in df.columns:
        if col_ref in df.columns:
            loc_var = df.columns.get_loc(col_ref)
            col2insert = df[col]
            df = df.drop([col], axis=1)
            if loc_var != 0:
                df.insert(loc_var - 1, col, col2insert)
            else:
                df.insert(0, col, col2insert)
    return df



# -*- coding: utf-8 -*-
# Copyright : INSEE, 2021

from functools import lru_cache
import pandas as pd


@lru_cache(maxsize=None)
def _get_nomenclature_agreg(file):

    df = pd.read_csv(file, sep=";", encoding="ISO-8859-1", dtype=str)

    for i in range(len(df.index)):
        for j in range(2, len(df.columns)):
            if pd.isna(df.iloc[i, j]):
                df.iloc[i, j] = df.iloc[i, j - 1]

    for j in range(len(df.columns)):
        for i in range(len(df.index)):
            if pd.isna(df.iloc[i, j]):
                df.iloc[i, j] = df.iloc[i - 1, j]

    def clean_TITLE(string):
        string_cleaned = string.replace("\x92", "'")
        return string_cleaned

    df = df.iloc[:, 1:10]
    df.columns = ["A10", "A21", "A38", "A64", "A88", "A129", "A138", "TITLE"]

    df.TITLE = df.TITLE.apply(clean_TITLE)

    return df



# -*- coding: utf-8 -*-
# Copyright : INSEE, 2021

import io
import logging
import xml.etree.ElementTree as ET

import pandas as pd

from ..utils.requests_session import PynseeAPISession
from ..utils.save_df import save_df
from ._get_dataset_list_internal import _get_dataset_list_internal


logger = logging.getLogger(__name__)


@save_df(day_lapse_max=30)
def get_dataset_list(update=False, silent=False):
    """Download a full INSEE's datasets list from BDM macroeconomic database

    Args:
        update (bool, optional): Set to True, to update manually the metadata
        stored locally on the computer. Defaults to False.

        silent (bool, optional): Set to True, to disable messages printed in log info

    Returns:
        DataFrame: a dataframe containing the list of datasets available

    Examples:
        >>> from pynsee.macrodata import get_dataset_list
        >>> insee_dataset = get_dataset_list()
    """

    try:

        INSEE_sdmx_link_dataflow = "https://bdm.insee.fr/series/sdmx/dataflow"
        INSEE_api_link_dataflow = (
            "https://api.insee.fr/series/BDM/dataflow/FR1/all"
        )

        with PynseeAPISession() as session:
            results = session.request_insee(
                sdmx_url=INSEE_sdmx_link_dataflow,
                api_url=INSEE_api_link_dataflow,
            )

        dataflow_file = io.BytesIO(results.content)
        root = ET.parse(dataflow_file).getroot()

        data = root[1][0]
        df = pd.DataFrame(
            [
                {
                    "id": next(iter(node.attrib.values())),
                    "Name.fr": node[1].text,
                    "Name.en": node[2].text,
                    "url": node[0][0][0].text,
                    "n_series": node[0][1][0].text,
                }
                for node in data
            ]
        )

        # clean up columns
        df = df.astype(str)

        df["n_series"] = df["n_series"].str.replace("\\D", "", regex=True)
        df["n_series"] = df["n_series"].astype("int")

        df = df[df["id"] != "SERIES_BDM"]

        df["Name.en"] = df["Name.en"].str.replace(
            "^\\n\\s{0,}", "", regex=True
        )
        df["Name.fr"] = df["Name.fr"].str.replace(
            "^\\n\\s{0,}", "", regex=True
        )
        df = df[df["Name.en"] != ""]
        df = df[df["Name.fr"] != ""]

    except Exception:
        df = _get_dataset_list_internal()

        logger.error(
            "Package's internal data has been used !\n"
            "Dataset list download failed !"
            "Please contact the package maintainer if this error persists !"
        )

    df = df.reset_index(drop=True)

    return df



from requests import RequestException

from ._find_wfs_closest_match import _find_wfs_closest_match
from ._get_geodata import _get_geodata
from .geofrdataframe import GeoFrDataFrame


def _get_geodata_with_backup(string: str) -> GeoFrDataFrame:
    try:
        df = _get_geodata(string, ignore_error=False)
    except RequestException as e:
        res = e.response

        if hasattr(res, "status_code") and res.status_code == 400:
            str_replacement = _find_wfs_closest_match(string)
            df = _get_geodata(str_replacement)
        else:
            raise e

    return df



# -*- coding: utf-8 -*-
# Copyright : INSEE, 2021

from functools import lru_cache
import os
import platformdirs

from pynsee.utils._get_temp_dir import _get_temp_dir
from pynsee.utils._hash import _hash

import logging

logger = logging.getLogger(__name__)


@lru_cache(maxsize=None)
def _create_insee_folder():

    try:

        # find local folder
        local_appdata_folder = platformdirs.user_cache_dir(
            "pynsee", ensure_exists=True
        )

        # create insee folder
        insee_folder = os.path.join(local_appdata_folder, "pynsee")
        os.makedirs(insee_folder, exist_ok=True)

        # test if saving a file is possible
        test_file = os.path.join(insee_folder, _hash("test_file"))
        with open(test_file, "w") as f:
            f.write("")
        # testing requires restricted rights on the machine

    except Exception:
        # if error temporary folder is returned
        insee_folder = _get_temp_dir()
        logger.warning(
            "pynsee folder could not be created in the appdata folder. "
            "Therefore, cache will only be handled in the current python "
            "session and you ARE responsible to clean the following temporary "
            "folder (at " + insee_folder + ") once you've finished!\n"
        )

    return insee_folder



# -*- coding: utf-8 -*-
# Copyright : INSEE, 2021

import re

import pandas as pd

from ..utils.save_df import save_df
from ._get_metadata import _get_dataset_metadata
from .get_dataset_list import get_dataset_list


@save_df(day_lapse_max=30)
def get_series_list(*datasets, update=False, silent=False):
    """Download an INSEE's series key list for one or several datasets from BDM macroeconomic database

    Args:
        datasets (str) : datasets should be among the datasets list provided by get_dataset_list()

        update (bool, optional): Set to True, to update manually the metadata
        stored locally on the computer. Defaults to False.

        silent (bool, optional): Set to True, to disable messages printed in log info

    Raises:
        ValueError: datasets should be among the datasets list provided by get_dataset_list()

    Returns:
        DataFrame: contains dimension columns, series keys, dataset name

    Notes:
        Some metadata is stored for 3 months locally on the computer. It is updated automatically

    Examples:
        >>> from pynsee.macrodata import get_dataset_list, get_series_list
        >>> dataset_list = get_dataset_list()
        >>> idbank_ipc = get_series_list('IPC-2015', 'CLIMAT-AFFAIRES')
    """
    insee_dataset = get_dataset_list(silent=silent, update=update)
    all_datasets = set(insee_dataset["id"])

    if len(datasets) == 1:
        if isinstance(datasets[0], list):
            datasets = datasets[0]

    wrong_datasets = [ds for ds in datasets if ds not in all_datasets]

    if wrong_datasets:
        raise ValueError(
            "The following are not datasets from INSEE: "
            f"{', '.join(wrong_datasets)}.\n"
            "Get a dataset list from the `get_dataset_list` function."
        )

    # get all datasets
    df = pd.concat(
        [
            _get_dataset_metadata(dt, update=update, silent=silent)
            for dt in datasets
        ],
        ignore_index=True,
    )

    # put label columns at the end and set names properly
    r = re.compile(".*_label_.*")
    column_all = df.columns.to_list()
    column_label = list(filter(r.match, column_all))
    column_other = [col for col in column_all if col not in column_label]
    new_column_order = column_other + column_label

    df = df.reindex(columns=new_column_order).rename(
        columns={"nomflow": "DATASET", "idbank": "IDBANK", "cleFlow": "KEY"}
    )

    df.columns = [col.replace("-", "_") for col in df.columns]

    return df



# -*- coding: utf-8 -*-
import logging

import pandas as pd

from pynsee.utils.requests_session import PynseeAPISession
from pynsee.geodata._geojson_parser import _geojson_parser
from pynsee.geodata._distance import _distance


logger = logging.getLogger(__name__)


def _set_global_var(args):
    global link0, list_bbox_full, session, crsPolygon0
    link0 = args[0]
    list_bbox_full = args[1]
    crsPolygon0 = args[2]

    session = PynseeAPISession()


def _get_data_with_bbox2(i):
    link = link0
    list_bbox = list_bbox_full[i]
    return _get_data_with_bbox(link, list_bbox, crsPolygon0)


def _get_data_with_bbox(link, list_bbox, crsPolygon="EPSG:4326"):
    # list_bbox = (5.0, 43.0, 6.0, 44.5)
    # bounds = [str(b) for b in list_bbox]
    # bounds = [bounds[1], bounds[0], bounds[3], bounds[2]]
    # BBOX= '&BBOX={}'.format(','.join(bounds))

    bounds = [str(b) for b in list_bbox]

    if crsPolygon == "EPSG:3857":
        bounds = [
            bounds[0],
            bounds[1],
            bounds[2],
            bounds[3],
            "urn:ogc:def:crs:" + crsPolygon,
        ]
    else:
        bounds = [
            bounds[1],
            bounds[0],
            bounds[3],
            bounds[2],
            "urn:ogc:def:crs:" + crsPolygon,
        ]

    BBOX = "&BBOX={}".format(",".join(bounds))

    link_query = link + BBOX

    if ("session" not in globals()) or ("session" not in locals()):
        session = PynseeAPISession()

    with session.get(link_query) as r:
        try:
            data_json = r.json()
        except Exception:
            logger.error(
                "The following query failed, some data might be missing !\n"
                f"{link_query}"
            )
            return pd.DataFrame()
        finally:
            session.close()

    if "features" in data_json.keys():
        json = data_json["features"]

        if len(json) > 0:
            if len(json) == 1000:
                # data limit reached
                # data searched relaunch with a smaller bbox
                # bbox area (longitude) divided by two

                width = _distance(
                    (list_bbox[0], list_bbox[1]), (list_bbox[2], list_bbox[1])
                )
                height = _distance(
                    (list_bbox[0], list_bbox[1]), (list_bbox[0], list_bbox[3])
                )

                if width > height:
                    list_bbox1 = [
                        list_bbox[0],
                        list_bbox[1],
                        (list_bbox[2] + list_bbox[0]) / 2,
                        list_bbox[3],
                    ]

                    list_bbox2 = [
                        (list_bbox[2] + list_bbox[0]) / 2,
                        list_bbox[1],
                        list_bbox[2],
                        list_bbox[3],
                    ]
                else:
                    list_bbox1 = [
                        list_bbox[0],
                        list_bbox[1],
                        list_bbox[2],
                        (list_bbox[1] + list_bbox[3]) / 2,
                    ]

                    list_bbox2 = [
                        list_bbox[0],
                        (list_bbox[1] + list_bbox[3]) / 2,
                        list_bbox[2],
                        list_bbox[3],
                    ]

                df1 = _get_data_with_bbox(link, list_bbox1)

                df2 = _get_data_with_bbox(link, list_bbox2)

                data_final = pd.concat([df1, df2]).reset_index(drop=True)
            else:
                # data limit not reached
                data_final = _geojson_parser(json).reset_index(drop=True)

        else:
            # no data found, return empty dataframe
            data_final = pd.DataFrame()
    else:
        # no data found, return empty dataframe
        data_final = pd.DataFrame()

    return data_final



import tempfile


def _import_options(
    tempdir: tempfile.TemporaryDirectory, caract: dict, filename: str
):
    """Internal to generate a dictionary of options
    required to import files

    Arguments:
        tempdir {tempfile.TemporaryDirectory} -- Temporary folder to use for unpacking
        caract {dict} -- Dictionary returned by `download_store_file`
        filename {str} -- Filename of the object that is going to be imported

    Returns:
        dict -- A dictionary listing options to control
         import with `pandas`
    """

    if caract["zip"] is True:
        file_archive = filename
        file_to_import = f"{tempdir}/{caract['fichier_donnees']}"
    else:
        file_archive = None
        file_to_import = filename

    import_args = {"file": file_to_import}

    if caract["type"] == "csv":
        import_args.update({"delim": caract["separateur"], "col_names": True})
        if "encoding" in list(caract.keys()):
            import_args.update({"locale": caract["encoding"]})
    elif caract["type"] in ["xls", "xlsx"]:
        import_args.update(
            {"path": file_to_import, "skip": caract["premiere_ligne"] - 1}
        )
        if "onglet" in list(caract.keys()):
            import_args.update({"sheet": caract["onglet"]})
        else:
            import_args.update({"sheet": 0})
        if "derniere_ligne" in list(caract.keys()):
            nmax_rows = caract["derniere_ligne"] - caract["premiere_ligne"]
            import_args.update({"n_max": nmax_rows})
        else:
            import_args.update({"n_max": None})

        if "valeurs_manquantes" in list(caract.keys()):
            import_args.update({"na": caract["valeurs_manquantes"]})
        else:
            import_args.update({"na": None})

    if "type_col" in list(caract.keys()):
        list_cols = caract["type_col"]
        for key, value in list_cols.items():
            if value == "character":
                list_cols[key] = "str"
            elif value == "integer":
                list_cols[key] = "int"
            elif value == "number":
                list_cols[key] = "float"
    else:
        list_cols = None

    import_args.update({"dtype": list_cols})
    try:
        encoding = caract["encoding"]
    except Exception:
        encoding = "Nan"
    import_args.update({"encoding": encoding})

    out_dict = {
        "file_archive": file_archive,
        "file_to_import": file_to_import,
        "import_args": import_args,
    }

    return out_dict



# -*- coding: utf-8 -*-
# Copyright : INSEE, 2021

import json
import logging

from functools import lru_cache
from typing import Dict

from pynsee.constants import CONFIG_FILE


logger = logging.getLogger(__name__)


def _get_credentials_from_configfile() -> Dict[str, str]:
    """
    Try to load credentials and proxy configuration from config file.

    Returns a dict containing at least an "sirene_key" entry.
    """
    key_dict: Dict[str, str] = {}

    try:
        with open(CONFIG_FILE, "r") as f:
            key_dict = json.load(f)

    except FileNotFoundError:
        # no credentials/config stored
        _missing_credentials()
        return key_dict

    return key_dict


@lru_cache(maxsize=None)
def _missing_credentials() -> None:
    logger.critical(
        "INSEE API credentials have not been found: please try to reuse "
        "pynsee.init_conn to save them locally.\n"
        "Otherwise, you can still use environment variables as follow:\n\n"
        "import os\n"
        "os.environ['sirene_key'] = 'my_sirene_key'"
    )



# -*- coding: utf-8 -*-
# Copyright : INSEE, 2021

from ._get_dimensions import _get_dimension_values


def _add_numeric_metadata(data):

    # data = get_dataset('CLIMAT-AFFAIRES')

    list_numeric_col_metadata = [
        "OBS_STATUS",
        "OBS_REV",
        "OBS_CONF",
        "OBS_TYPE",
        "OBS_QUAL",
    ]

    for col in list_numeric_col_metadata:
        cl_col = "CL_" + col
        df = _get_dimension_values(cl_col, silent=False)

        df = df[df["id"] != cl_col]

        # rename columns
        df.columns = [col, col + "_label_fr", col + "_label_en"]

        if col in data.columns:
            data = data.merge(df, on=col, how="left")

    return data



import tempfile

from pandas import DataFrame

from pynsee.download._download_store_file import _download_store_file
from pynsee.download._load_data_from_schema import _load_data_from_schema

from pynsee.utils.save_df import save_df


@save_df(day_lapse_max=90)
def download_file(id, variables=None, update=False, silent=False) -> DataFrame:
    """
    User level function to download files from insee.fr

    Note: in the case of vintaged ids (ie. TAG_COM_2025), it should be safe to
    use a _LATEST tag instead if you just want the latest available dataset
    (ie TAG_COM_LATEST). In case of side-effects, pay attention to the log
    entries and revert to the vintaged id.

    Args:
        id (str): file id, check get_file_list to have a full list of available files

        variables (list): a list of variables to load from the data file, use
          get_column_metadata function to have the full list

        update (bool, optional): Trigger an update, otherwise locally saved data is
          used. Defaults to False.

        silent (bool, optional): Set to True, to disable messages printed in log info

    Returns:
        Returns the request dataframe as a pandas object

    Examples:
        >>> from pynsee.download import download_file
        >>> df = download_file("TAG_COM_LATEST")

    """

    with tempfile.TemporaryDirectory() as tmpdir:

        dwn = _download_store_file(tmpdir, id)
        df = _load_data_from_schema(dwn, variables=variables)

    return df



# -*- coding: utf-8 -*-
# Copyright : INSEE, 2021

from pynsee.macrodata.get_series import get_series


def get_series_title(series):
    """Get French and English titles of a list of series (idbanks)

    Args:
        series (list): a list of series (idbanks)

    Examples:
        >>> from pynsee import get_series_list, get_series_title
        >>> series = get_series_list("CLIMAT-AFFAIRES")
        >>> series = series.loc[:3, "IDBANK"].to_list()
        >>> titles = get_series_title(series)
    """
    data = get_series(series, firstNObservations=1, metadata=False)

    data = data[["IDBANK", "TITLE_FR", "TITLE_EN"]].reset_index(drop=True)

    return data



# -*- coding: utf-8 -*-
# Copyright : INSEE, 2021

import io
from functools import lru_cache
import pandas as pd
import xml.etree.ElementTree as ET
from tqdm import trange

from pynsee.utils.requests_session import PynseeAPISession


@lru_cache(maxsize=None)
def _get_geo_list_simple(geo, date=None, progress_bar=False):
    api_url = "https://api.insee.fr/metadonnees/geo/" + geo
    if date:
        api_url += f"?date={date}"

    with PynseeAPISession() as session:
        results = session.request_insee(api_url=api_url)

    raw_data_file = io.BytesIO(results.content)

    root = ET.parse(raw_data_file).getroot()

    n_variable = len(root)

    list_data_geo = []

    if progress_bar is True:
        geo_range = trange(n_variable, desc="Getting %s" % geo, leave=False)
    else:
        geo_range = range(n_variable)

    for igeo in geo_range:
        n_var = len(root[igeo])

        dict_geo = {}

        for ivar in range(n_var):
            dict_geo[root[igeo][ivar].tag] = root[igeo][ivar].text

        dict_geo = {**dict_geo, **root[igeo].attrib}

        data_geo = pd.DataFrame(dict_geo, index=[0])

        list_data_geo.append(data_geo)

    df_geo = pd.concat(list_data_geo, ignore_index=True)
    return df_geo



from pynsee.download.get_file_list import get_file_list
from pynsee.download.download_file import download_file
from pynsee.download.get_column_metadata import get_column_metadata

__all__ = ["get_file_list", "download_file", "get_column_metadata"]



# -*- coding: utf-8 -*-
# Copyright : INSEE, 2021

from functools import lru_cache
import unidecode
import pandas as pd

from ._get_idbank_internal_data import _get_idbank_internal_data


@lru_cache(maxsize=None)
def _get_idbank_internal_data_harmonized():

    idbank_list = _get_idbank_internal_data()

    col_selected = ["DATASET", "IDBANK", "KEY", "TITLE_FR", "TITLE_EN"]
    idbank_col_selected = idbank_list.columns[
        idbank_list.columns.isin(col_selected)
    ]
    idbank_list = idbank_list[idbank_col_selected]

    # all titles in lower case
    title_fr_lower = pd.Series(
        [str(title).lower() for title in idbank_list["TITLE_FR"]]
    )
    idbank_list = idbank_list.assign(title_fr_lower=title_fr_lower.values)

    title_en_lower = pd.Series(
        [str(title).lower() for title in idbank_list["TITLE_EN"]]
    )
    idbank_list = idbank_list.assign(title_en_lower=title_en_lower.values)

    # create column without accent
    title_fr_no_accent = [
        unidecode.unidecode(str(title))
        for title in idbank_list["title_fr_lower"]
    ]
    title_fr_no_accent = pd.Series(title_fr_no_accent)
    idbank_list = idbank_list.assign(
        title_fr_no_accent=title_fr_no_accent.values
    )

    return idbank_list



import io
import zipfile
import importlib
import pandas as pd

from pynsee.utils.save_df import save_df


@save_df(day_lapse_max=90)
def _get_dataset_list_internal(silent=False):
    """
    silent (bool, optional): Set to True, to disable messages printed in log info
    """

    try:
        pkg_macrodata = importlib.resources.files(__name__)
        zip_file = str(pkg_macrodata) + "/data/dataset_list_internal.zip"
    except Exception:
        import pkg_resources

        zip_file = pkg_resources.resource_stream(
            __name__, "data/dataset_list_internal.zip"
        )

    with zipfile.ZipFile(zip_file, "r") as zip_ref:
        dataset_file = io.BytesIO(zip_ref.read("dataset_list_internal.csv"))

    dataset_list = pd.read_csv(
        dataset_file, encoding="utf-8", quotechar='"', sep=",", dtype=str
    )

    col = "Unnamed: 0"
    if col in dataset_list.columns:
        dataset_list = dataset_list.drop(columns={col})

    return dataset_list



# -*- coding: utf-8 -*-

import io
from functools import lru_cache
from pynsee.utils.requests_session import PynseeAPISession


@lru_cache(maxsize=None)
def _get_capabilities(key="", version="1.0.0", service="wmts") -> io.BytesIO:

    service_upper = service.upper()

    link = (
        f"https://data.geopf.fr/{service}?SERVICE={service_upper}"
        f"&VERSION={version}&REQUEST=GetCapabilities"
    )

    with PynseeAPISession() as session:

        results = session.get(link, verify=False)
        raw_data_file = io.BytesIO(results.content)

    return raw_data_file



import json
import os

import requests
from pynsee.utils._create_insee_folder import _create_insee_folder
from pynsee.utils._hash import _hash
from pynsee.utils._warning_cached_data import _warning_cached_data


def _get_location_openstreetmap(
    query: str, session: requests.Session, update: bool = False
) -> tuple[float, float, str, str, float]:
    """
    Query a location using Nominatim and cache it on local disk.

    Parameters
    ----------
    query : str
        query, fo instance '2+Rue+DES+FRERES+BERTRAND+69200+VENISSIEUX+FRANCE'
    session : requests.Session
        requests.Session object used for the connection.
    update : bool, optional
        If False, will first try to restore previous data from disk. The
        default is False.

    Returns
    -------
    tuple[float, float, str, str, float]
        latitude, longitude, location category, location type, importance

    """
    api_link = (
        "https://nominatim.openstreetmap.org/search.php?"
        f"q={query}&format=jsonv2&limit=1"
    )

    insee_folder = _create_insee_folder()
    filename = os.path.join(insee_folder, f"{_hash(api_link)}.json")

    if update or not os.path.isfile(filename):
        results = session.get(api_link)
        data = results.json()

        with open(filename, "w") as f:
            json.dump(data, f)
    else:
        with open(filename, "r") as f:
            data = json.load(f)

        _warning_cached_data(filename)

    data = data[0]
    return (
        float(data["lat"]),
        float(data["lon"]),
        data["category"],
        data["type"],
        data["importance"],
    )



# -*- coding: utf-8 -*-
# Copyright : INSEE, 2021

from functools import lru_cache
from pynsee.utils.requests_session import PynseeAPISession
from pynsee.utils._create_insee_folder import _create_insee_folder
from pynsee.utils._make_dataframe_from_dict import _make_dataframe_from_dict

import zipfile
import os
import importlib
import pandas as pd

import logging

logger = logging.getLogger(__name__)


@lru_cache(maxsize=None)
def _warning_definition_internal_data():
    logger.info(
        "Internal package data has been used !\n"
        "If some data is missing, please use get_definition !"
    )


@lru_cache(maxsize=None)
def get_definition_list():
    """Get a list of concept definitions

    Examples:
        >>> from pynsee.metadata import get_definition_list
        >>> definition = get_definition_list()
    """

    insee_folder = _create_insee_folder()

    insee_folder_local_def = insee_folder + "/" + "definition"

    os.makedirs(insee_folder_local_def, exist_ok=True)

    list_expected_files = ["all_definitions.csv"]

    list_expected_files = [
        insee_folder + "/definition/" + f for f in list_expected_files
    ]

    list_available_file = [not os.path.exists(f) for f in list_expected_files]

    # unzipping raw files
    if any(list_available_file):
        try:
            pkg_path = importlib.resources.files(__name__)
            zip_file = str(pkg_path) + "/data/definition.zip"
        except Exception:
            import pkg_resources

            zip_file = pkg_resources.resource_stream(
                __name__, "data/definition.zip"
            )

        with zipfile.ZipFile(zip_file, "r") as zip_ref:
            zip_ref.extractall(insee_folder)

    link = "https://api.insee.fr/metadonnees/concepts/definitions"

    with PynseeAPISession() as session:
        request = session.request_insee(
            api_url=link, file_format="application/json"
        )

    data_request = request.json()

    list_data = []

    for i in range(len(data_request)):
        df = _make_dataframe_from_dict(data_request[i])
        df = df.iloc[:, 0:3].reset_index(drop=True).drop_duplicates()
        list_data.append(df)

    data = pd.concat(list_data, axis=0)
    data = data.reset_index(drop=True)
    data.columns = ["ID", "URI", "TITLE_FR"]

    if os.path.exists(list_expected_files[0]):
        all_data = pd.read_csv(list_expected_files[0])
        all_data = all_data.iloc[:, 1:10]
        all_data = all_data.drop(columns={"URI", "TITLE_FR"})
        data = data.merge(all_data, on="ID", how="left")

    _warning_definition_internal_data()

    return data



# -*- coding: utf-8 -*-
# Copyright : INSEE, 2021

from pynsee.macrodata.get_dataset_list import get_dataset_list
from pynsee.macrodata.get_series_list import get_series_list
from pynsee.macrodata.get_series_title import get_series_title
from pynsee.macrodata.search_macrodata import search_macrodata

from tqdm import trange
import os
import pandas as pd


def _build_series_list(dt=["CNA-2014-ERE"]):
    #
    # FIRST change the link, file name and separator in _download_idbank_list
    # THEN build the package and try to use the following to make a new idbank_list file
    # FINALLY put the file in the zip file in macrodata/data
    #

    #
    # SET dt = None TO BUILD THE FULL DATAFRAME
    #

    os.environ["pynsee_use_sdmx"] = "True"

    if dt is None:
        dt = get_dataset_list()
        dt = dt.id.to_list()

    list_dt = []

    for d in trange(len(dt)):
        dataset = dt[d]
        df = get_series_list(dataset, update=True)
        list_dt.append(df)

    series_list = pd.concat(list_dt)
    # series_list.to_csv("pynsee_series_all.csv")
    # series_list = pd.read_csv("../pynsee_series_all.csv", dtype=str)
    # col = 'Unnamed: 0'
    # series_list = series_list.drop(columns = {col})

    old_series = search_macrodata()
    old_series = old_series.drop(columns={"KEY"})
    series_list_short = series_list[["DATASET", "IDBANK", "KEY"]]

    series_list_new = series_list_short.merge(
        old_series, on=["DATASET", "IDBANK"], how="left"
    )

    series_list_new_title_missing = series_list_new[
        pd.isna(series_list_new["TITLE_FR"])
    ]
    series_list_new_title_missing = series_list_new_title_missing.drop(
        columns={"TITLE_FR", "TITLE_EN"}
    )

    list_series_title_missing = series_list_new_title_missing.IDBANK.to_list()

    if len(list_series_title_missing) > 0:
        titles = get_series_title(list_series_title_missing)

        # !!!!!!!!!!!!!!!!!!!!!!
        # SERIES IN IDBANK FILE NOT AVAILABLE IN BDM DATABASE
        # !!!!!!!!!!!!!!!!!!!!!!
        # series_not_found = [l for l in list_series_title_missing if l not in titles.IDBANK.to_list()]

        if len(titles.index) > 0:
            titles = titles.rename(
                columns={"TITLE_FR": "TITLE_FR2", "TITLE_EN": "TITLE_EN2"}
            )

            series_list_new2 = series_list_new.merge(
                titles, on="IDBANK", how="left"
            )

            for i in range(len(series_list_new2.index)):
                tt_fr = series_list_new2.loc[i, "TITLE_FR"]
                tt_en = series_list_new2.loc[i, "TITLE_EN"]
                if pd.isna(tt_fr):
                    series_list_new2.loc[i, "TITLE_FR"] = series_list_new2.loc[
                        i, "TITLE_FR2"
                    ]

                if pd.isna(tt_en):
                    series_list_new2.loc[i, "TITLE_EN"] = series_list_new2.loc[
                        i, "TITLE_EN2"
                    ]

            series_list_new2 = series_list_new2.drop(
                columns={"TITLE_FR2", "TITLE_EN2"}
            )
            # series_not_found2 = series_list_new2[pd.isna(series_list_new2['TITLE_FR'])]
            series_list_new2 = series_list_new2.dropna(subset={"TITLE_FR"})
            series_list_new2 = series_list_new2[
                ["IDBANK", "TITLE_FR", "TITLE_EN"]
            ]

            series_list2 = series_list.merge(
                series_list_new2, on="IDBANK", how="right"
            )

            # !!!!!!!!!
            # CREATE FILE FOR PACKAGE INTERNAL DATA
            # !!!!!!!!!
            # series_list2.to_csv("idbank_list_internal.csv", encoding = "utf-8", quotechar='"', sep=',')

        os.environ["pynsee_use_sdmx"] = "False"
        return series_list2
    else:
        os.environ["pynsee_use_sdmx"] = "False"
        return old_series



from pynsee.geodata import *
from pynsee.macrodata import *
from pynsee.localdata import *
from pynsee.sirene import *
from pynsee.metadata import *
from pynsee.utils import *
from pynsee.download import *

__version__ = "0.2.3"



from .geofrdataframe import GeoFrDataFrame


def get_geom(gdf: GeoFrDataFrame):
    """Extract a shape (Polygon, Point ...) from a GeoFrDataFrame

    .. deprecated:: 0.2.0

        Use :meth:`GeoFrDataFrame.geometry` instead.

    Examples:
        >>> from pynsee.geodata import get_geodata_list, get_geodata
        >>> #
        >>> # Get a list of geographical limits of French administrative areas from IGN API
        >>> geodata_list = get_geodata_list()
        >>> #
        >>> # Get geographical limits of departments
        >>> df = get_geodata('ADMINEXPRESS-COG-CARTO.LATEST:departement')
        >>> #
        >>> # Extract a polygon from the GeoDataframe
        >>> geo = df.get_geom()
    """
    return gdf.get_geom()



import pandas as pd
from functools import lru_cache

from pynsee.download._get_dict_data_source import _get_dict_data_source
from pynsee.utils._move_col_after import _move_col_before


import logging

logger = logging.getLogger(__name__)


@lru_cache(maxsize=None)
def get_file_list():
    """Download a list of files available on insee.fr

    Returns:
        Returns the requested dataframe as a pandas object

    Notes:
        pynsee.download's metadata rely on volunteering contributors and their manual updates.
        get_file_list does not provide data from official Insee's metadata API.
        Consequently, please report any issue

    Examples:
        >>> from pynsee.download import get_file_list
        >>> insee_file_list = get_file_list()

    """

    df = pd.DataFrame(_get_dict_data_source()).T

    df["id"] = df.index
    df = df.reset_index(drop=True)
    df = _move_col_before(df, "id", "nom")

    rename_col_dict = {
        "nom": "name",
        "libelle": "label",
        "lien": "link",
        "fichier_donnees": "data_file",
        "onglet": "tab",
        "premiere_ligne": "first_row",
        "fichier_meta": "meta_file",
        "separateur": "separator",
        "derniere_ligne": "last_row",
        "valeurs_manquantes": "missing_value",
        "disponible": "available",
    }
    df = df.rename(columns=rename_col_dict)

    df = df[~df.link.str.contains("https://api.insee.fr")]

    warning_metadata_download()

    return df


@lru_cache(maxsize=None)
def warning_metadata_download():
    logger.info(
        "pynsee.download's metadata rely on volunteering contributors and "
        "their manual updates. "
        "get_file_list does not provide data from official Insee's metadata "
        "API\nConsequently, please report any issue"
    )



import pandas as pd
from tqdm import trange

from ..utils.save_df import save_df
from .get_dataset_list import get_dataset_list
from ._get_metadata import _get_dataset_metadata


@save_df(day_lapse_max=90)
def _load_dataset_data(update=False, silent=False):

    list_dataset = list(get_dataset_list(silent=silent).id.unique())
    list_metadata = []

    for dt in trange(len(list_dataset), desc="Metadata download"):
        dataset = list_dataset[dt]
        metadata = _get_dataset_metadata(dataset, silent=silent)
        list_metadata += [metadata]

    return pd.concat(list_metadata)



# -*- coding: utf-8 -*-
from math import sin, cos, sqrt, atan2, radians


def _distance(point1, point2):
    # see https://stackoverflow.com/questions/19412462/getting-distance-between-two-points-based-on-latitude-longitude

    # approximate radius of earth in km
    R = 6373.0

    lat1 = radians(point1[1])
    lon1 = radians(point1[0])
    lat2 = radians(point2[1])
    lon2 = radians(point2[0])

    dlon = lon2 - lon1
    dlat = lat2 - lat1

    a = sin(dlat / 2) ** 2 + cos(lat1) * cos(lat2) * sin(dlon / 2) ** 2
    c = 2 * atan2(sqrt(a), sqrt(1 - a))

    distance = R * c

    return distance



# -*- coding: utf-8 -*-
# Copyright : INSEE, 2021

import io
from functools import lru_cache
import xml.etree.ElementTree as ET
import pandas as pd
import re
import datetime
import warnings

from pynsee.utils.requests_session import PynseeAPISession


@lru_cache(maxsize=None)
def get_last_release():
    """Get the datasets from BDM macroeconomic database released in the last 30 days

    Examples
        >>> from pynsee.macrodata import get_last_release
        >>> dataset_released = get_last_release()
    """
    link = "https://bdm.insee.fr/series/sdmx/rss/donnees"
    try:
        with PynseeAPISession() as session:
            request = session.request_insee(sdmx_url=link)

        file = io.BytesIO(request.content)

        root = ET.parse(file).getroot()

        list_data = []

        for i in range(len(root[0])):
            if root[0][i].tag == "item":
                dico = {}

                for j in range(len(root[0][i])):
                    if root[0][i][j].tag == "pubDate":
                        date = datetime.datetime.strptime(
                            root[0][i][j].text, "%a, %d %b %Y %H:%M:%S GMT"
                        )
                        dico["pubDate"] = date
                    else:
                        dico[root[0][i][j].tag] = root[0][i][j].text

                    if root[0][i][j].tag == "title":
                        string = re.search("\\[.*\\]", root[0][i][j].text)
                        if string:
                            string_selected = (
                                string.group(0)
                                .replace("[", "")
                                .replace("]", "")
                            )
                        else:
                            string_selected = string
                        dico["dataset"] = string_selected

                df = pd.DataFrame(dico, index=[0])
                list_data.append(df)

        data = pd.concat(list_data)

    except Exception as e:
        warnings.warn(str(e))
        data = pd.DataFrame()

    return data



# -*- coding: utf-8 -*-
# Copyright : INSEE, 2021

import functools
import logging

import pandas as pd

from ..utils.save_df import save_df
from ._download_idbanks import _download_idbank_list
from ._get_dimensions import _get_dataset_dimension, _get_dimension_values
from ._get_idbank_internal_data import _get_idbank_internal_data


logger = logging.getLogger(__name__)


@functools.lru_cache(maxsize=None)
def _warning_error():
    logger.error(
        "Package's internal data has been used !\n"
        "Idbank file download failed, have a look at the following page "
        "and find the new link !\n"
        "https://www.insee.fr/en/information/2868055\n\n"
        "You may change the downloaded file changing the following "
        "environment variable !\n"
        "import os; os.environ['pynsee_idbank_file'] = 'my_new_idbank_file'\n"
        "Please contact the package maintainer if this error persists !"
    )


@save_df(day_lapse_max=90)
def _get_dataset_metadata(
    dataset, update=False, silent=False, insee_date_test=None
):
    """Get dataset metadata; use local data as backup if server fails."""
    try:
        idbank_list_dataset = _get_dataset_metadata_core(
            dataset=dataset, update=update, silent=silent
        )
    except Exception:
        # if the download of the idbank file and the build of the metadata fail
        # package's internal data is provided to the user, should be
        # exceptional, used as a backup
        _warning_error()

        idbank_list_dataset = _get_idbank_internal_data(update=update)
        idbank_list_dataset = idbank_list_dataset[
            idbank_list_dataset["DATASET"] == dataset
        ]

        # drop the columns where all elements are NaN
        idbank_list_dataset = idbank_list_dataset.dropna(axis=1, how="all")
        idbank_list_dataset = idbank_list_dataset.reset_index(drop=True)

    return idbank_list_dataset


@save_df(day_lapse_max=90)
def _get_dataset_metadata_core(dataset, update=False, silent=False):
    """Get metadata from idbanks"""
    idbank_list = _download_idbank_list(update=update, silent=silent)

    # get dataset's dimensions
    dataset_dimension = _get_dataset_dimension(
        dataset, update=update, silent=silent
    ).reset_index(drop=True)

    # select only the idbanks corresponding to the dataset
    idbank_list_dataset = idbank_list[
        idbank_list["nomflow"] == dataset
    ].reset_index(drop=True)

    # split the cleflow column with the dot as a separator
    df_cleflow_splitted = idbank_list_dataset.cleFlow.str.split("\\.").tolist()

    # map the original to the new column names
    convertor = {
        row.local_representation[3:]: row.dimension
        for _, row in dataset_dimension.iterrows()
    }

    colnames = idbank_list_dataset.iloc[0].list_var.split(".")

    new_columns = [convertor[c] for c in colnames]

    # make a dataframe from the splitted cleflow using the new column names
    df_cleflow_splitted = pd.DataFrame(
        df_cleflow_splitted, columns=new_columns
    )

    # join the splitted cleflow dataframe with the former idbank list
    idbank_list_dataset = pd.concat(
        [idbank_list_dataset, df_cleflow_splitted], axis=1
    )

    for irow, dim_id in dataset_dimension["dimension"].items():
        if dim_id not in idbank_list_dataset.columns:
            continue

        dim_local_rep = dataset_dimension["local_representation"].iloc[irow]

        # get dimension values
        dim_values = _get_dimension_values(
            dim_local_rep, update=update, silent=silent
        )

        # drop dimension label
        dim_values = dim_values[dim_values["id"] != dim_local_rep]

        # rename columns
        dim_values.columns = [
            dim_id,
            dim_id + "_label_fr",
            dim_id + "_label_en",
        ]
        idbank_list_dataset = idbank_list_dataset.merge(
            dim_values, on=dim_id, how="left"
        )

    return idbank_list_dataset.drop(columns=["list_var"])



from tqdm import tqdm


from pynsee.utils.requests_session import PynseeAPISession


def _download_pb(url: str, fname: str, total: int = None):
    """Useful function to get request with a progress bar

    Borrowed from
    https://gist.github.com/yanqd0/c13ed29e29432e3cf3e7c38467f42f51

    Arguments:
        url {str} -- URL for the source file
        fname {str} -- Destination where data will be written
    """

    with PynseeAPISession() as s:
        resp = s.get(url, stream=True, verify=False)

    if total is None:
        total = int(resp.headers.get("content-length", 0))

    with (
        open(fname, "wb") as file,
        tqdm(
            desc="Downloading: ",
            total=total,
            unit="iB",
            unit_scale=True,
            unit_divisor=1024,
        ) as obj,
    ):
        for data in resp.iter_content(chunk_size=1024):
            size = file.write(data)
            obj.update(size)



# -*- coding: utf-8 -*-
# Copyright : INSEE, 2021


from functools import lru_cache
import logging
import xml.etree.ElementTree as ET

import pandas as pd

from pynsee.utils.requests_session import PynseeAPISession


logger = logging.getLogger(__name__)


def _set_date(df: pd.DataFrame) -> pd.DataFrame:
    """
    Parse dates on a dataframe issued by macrodata module.

    Adds a "DATE" first column (computed from the "TIME_PERIOD" original
    column) and sort the dataframe according to it.

    Internal note: leave this as a separate function to to improve tests

    coverage.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame containing at least two columns: "FREQ" and "TIME_PERIOD".
        The FREQ column should contain only one homogeneous value.

    Returns
    -------
    df : pd.DataFrame
    """

    freqs = df.FREQ.unique()
    if "DATE" not in df.columns:
        df["DATE"] = None

    for freq in freqs:
        ix = df[df.FREQ == freq].index
        time = df.loc[ix, "TIME_PERIOD"]
        if freq == "M":
            df.loc[ix, "DATE"] = time + "-01"

        elif freq == "A":
            df.loc[ix, "DATE"] = time + "-01-01"

        elif freq in {"T", "S", "B"}:

            if freq == "T":
                pat = "(-Q[1234])"
                repl = {
                    "-Q1": "-01-01",
                    "-Q2": "-04-01",
                    "-Q3": "-07-01",
                    "-Q4": "-10-01",
                }
            elif freq == "S":
                pat = "(-S[12])"
                repl = {"-S1": "-01-01", "-S2": "-07-01"}
            elif freq == "B":
                pat = "(-B[123456])"
                repl = {
                    "-B1": "-01-01",
                    "-B2": "-03-01",
                    "-B3": "-05-01",
                    "-B4": "-07-01",
                    "-B5": "-09-01",
                    "-B6": "-11-01",
                }

            df.loc[ix, "DATE"] = time.str[:-3] + time.str.extract(
                pat, expand=False
            ).map(repl)
        else:
            df.loc[ix, "DATE"] = time

        if freq in {"M", "A", "S", "T", "B"}:
            df.loc[ix, "DATE"] = pd.to_datetime(
                df.loc[ix, "DATE"], format="%Y-%m-%d"
            )

    if "DATE" in df.columns:
        # place DATE column in the first position and sort
        df[["DATE"] + [c for c in df.columns if c not in ["DATE"]]]
        df = df.sort_values(["IDBANK", "DATE"]).reset_index(drop=True)

    return df


@lru_cache(maxsize=None)
def _get_insee(api_query, sdmx_query, step="1/1"):

    with PynseeAPISession() as session:
        results = session.request_insee(sdmx_url=sdmx_query, api_url=api_query)

    # parse the xml file
    root = ET.fromstring(results.content)

    obs_series = []
    series = root.findall(".//Series")
    for serie in series:
        dict_attr = serie.attrib
        obs = [obs.attrib for obs in serie.iterfind("./")]
        [x.update(dict_attr) for x in obs]
        obs_series += obs
    data = pd.DataFrame(obs_series)

    data = _set_date(data)

    # harmonise column names
    colnames = data.columns

    def replace_hyphen(x):
        return str(x).replace("-", "_")

    newcolnames = list(map(replace_hyphen, colnames))
    data.columns = newcolnames

    if "OBS_VALUE" in data.columns:
        data["OBS_VALUE"] = data["OBS_VALUE"].apply(
            pd.to_numeric, errors="coerce"
        )

    logger.debug("Data has been cached")

    return data



from functools import lru_cache
import logging
from typing import Union
import warnings

import geopandas as gpd
import numpy as np
import pandas as pd
from shapely.errors import ShapelyDeprecationWarning
from tqdm.auto import tqdm
from unidecode import unidecode

from pynsee.geodata import GeoFrDataFrame
from pynsee.utils.requests_session import PynseeAPISession
from ._get_location_openstreetmap import (
    _get_location_openstreetmap,
)

logger = logging.getLogger(__name__)

tqdm.pandas(desc="Getting location")


@lru_cache(maxsize=None)
def _warning_get_location():
    logger.warning(
        "For at least one point, exact location has not been found, city "
        "location has been given instead"
    )


@lru_cache(maxsize=None)
def _warning_OSM():
    logger.info(
        "This function returns data made available by OpenStreetMap and its "
        "contributors.\n"
        "Please comply with Openstreetmap's Copyright and ODbL Licence"
    )


CONFIG_ADDRESSES = {
    "address": [
        "numeroVoieEtablissement",
        "typeVoieEtablissementLibelle",
        "libelleVoieEtablissement",
        "codePostalEtablissement",
        "libelleCommuneEtablissement",
    ],
    "address_backup": [
        "codePostalEtablissement",
        "libelleCommuneEtablissement",
    ],
}

NOMINATIM_RESULTS = ["latitude", "longitude", "category", "type", "importance"]


def clean_cities(s: pd.Series) -> pd.Series:
    "Do some precleaning for cities names"
    s = (
        s.str.upper()
        .apply(unidecode)
        # Neuville-Housset (La) -> Neuville-Housset:
        .str.replace(r" \(.*\)$", "", regex=True)
        .str.replace(r"(^|\s)(ST)\s", " SAINT ", regex=True)
        .str.replace(r"(^|\s)(STE)\s", " SAINTE ", regex=True)
        # PARIS 4, PARIS 1er etc., but do not capture "LE TAMPON 14EME KM" from La Réunion noqa: E501
        .str.replace(r"[0-9]+(ER?|EME)?(\s(?!KM(\s|$)|$)|$)", "", regex=True)
        .str.strip()
        .str.replace(r" ?CEDEX$", "", regex=True)
    )
    return s


class SireneDataFrame(pd.DataFrame):
    """Class for handling dataframes built from INSEE SIRENE API's data"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return SireneDataFrame

    def get_location(
        self, update=False
    ) -> Union[GeoFrDataFrame, "SireneDataFrame"]:
        """
        Get latitude and longitude from OpenStreetMap, add geometry column and
        turn ``SireneDataframe`` into ``GeoFrDataFrame``.

        Args:
            update (bool, optional): data is saved locally, set update=True to
                trigger an update. Defaults to False.

        Notes:
            If it fails to find the exact location, by default it returns the
            location of the city. Whether the exact location has been found or
            not is encoded in the `exact_location` column of the new
            ``GeoFrDataFrame``.

        Examples:
            >>> from pynsee.metadata import get_activity_list
            >>> from pynsee.sirene import search_sirene
            >>> #
            >>> #  Get activity list
            >>> naf5 = get_activity_list('NAF5')
            >>> #
            >>> # Get alive legal entities belonging to the automotive industry
            >>> df = search_sirene(variable = ["activitePrincipaleEtablissement"],
            >>>                    pattern = ['29.10Z'], kind = 'siret')
            >>> #
            >>> # Keep businesses with more than 100 employees
            >>> df = df.loc[df['effectifsMinEtablissement'] > 100]
            >>> df = df.reset_index(drop=True)
            >>> #
            >>> # Get location
            >>> df = df.get_location()
        """
        _warning_OSM()

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", category=ShapelyDeprecationWarning
            )

            df = self.reset_index(drop=True)

            list_col = [
                "siret",
                "numeroVoieEtablissement",
                "typeVoieEtablissementLibelle",
                "libelleVoieEtablissement",
                "codePostalEtablissement",
                "libelleCommuneEtablissement",
            ]

            if set(list_col).issubset(df.columns):

                addresses = df[CONFIG_ADDRESSES["address"]].fillna("")

                city = "libelleCommuneEtablissement"
                addresses[city] = clean_cities(addresses[city])
                for field in [
                    city,
                    "libelleVoieEtablissement",
                    "typeVoieEtablissementLibelle",
                ]:
                    addresses[field] = addresses[field].str.replace(
                        " (D|L) ", r" \1'", case=False, regex=True
                    )

                for field, config in CONFIG_ADDRESSES.items():
                    addresses[field] = (
                        pd.Series(addresses[config].values.tolist())
                        .str.join(" ")
                        .str.replace(" +", "+", regex=True)
                        .str.strip()
                    )
                    ix = addresses[addresses[field] != ""].index
                    addresses.loc[ix, field] += "+FRANCE"
                    ix = addresses[addresses[field] == ""].index
                    addresses.loc[ix, field] = np.nan

                with PynseeAPISession() as session:

                    def query_func(x):
                        try:
                            return _get_location_openstreetmap(
                                x,
                                session=session,
                                update=update,
                            )
                        except IndexError:
                            return None

                    ix = addresses.index
                    for field in CONFIG_ADDRESSES:
                        sample = (
                            addresses.loc[ix, [field]]
                            .drop_duplicates()
                            .dropna()
                        )

                        # Use TQDM extension
                        # see https://tqdm.github.io/docs/tqdm/#pandas
                        sample["result"] = sample[field].progress_apply(
                            query_func
                        )
                        if "result" in addresses:
                            # simple hack to rename column for second config
                            sample = sample.rename(
                                {"result": "result2"}, axis=1
                            )
                        addresses = addresses.merge(
                            sample, on=field, how="left"
                        )
                        ix = addresses[addresses["result"].isnull()].index
                addresses["exact_location"] = ~addresses["result"].isnull()
                addresses["result"] = addresses["result"].combine_first(
                    addresses["result2"]
                )
                addresses = addresses.drop("result2", axis=1)
                addresses["exact_location"] = (
                    addresses["exact_location"].fillna(1).astype("bool")
                )

                def extract_data(tup):
                    try:
                        return dict(zip(NOMINATIM_RESULTS, tup))
                    except TypeError:
                        return {}

                results = pd.DataFrame(
                    addresses["result"].apply(extract_data).values.tolist()
                )
                addresses = addresses.drop("result", axis=1).join(results)

                addresses = addresses[["exact_location"] + NOMINATIM_RESULTS]
                addresses["geometry"] = gpd.points_from_xy(
                    x=addresses["longitude"], y=addresses["latitude"], crs=4326
                )
                addresses = addresses.drop(["latitude", "longitude"], axis=1)
                return GeoFrDataFrame(
                    self.join(addresses, how="left"), crs=4326
                )

            return df



# -*- coding: utf-8 -*-
# Copyright : INSEE, 2021

import pandas as pd
from functools import lru_cache
from tqdm import trange

from pynsee.utils.requests_session import PynseeAPISession
from pynsee.utils.save_df import save_df


@save_df(day_lapse_max=30)
def get_legal_entity(codes, print_err_msg=True, update=False, silent=False):
    """Get legal entities labels

    Args:
        codes (list): list of legal entities code of 2 or 4 characters

        update (bool, optional): Trigger an update, otherwise locally saved data is used. Defaults to False

        silent (bool, optional): Set to True, to disable messages printed in log info

    Examples:
        >>> from pynsee.metadata import get_legal_entity
        >>> legal_entity = get_legal_entity(codes = ['5599', '83'])
    """

    list_data = []

    for c in trange(len(codes), desc="Getting legal entities"):
        # c = '5599'
        code = codes[c]
        try:
            data = _get_one_legal_entity(code, print_err_msg=print_err_msg)
            list_data.append(data)
        except Exception:
            pass

    data_final = pd.concat(list_data).reset_index(drop=True)

    data_final = data_final.rename(columns={"intitule": "title"})

    return data_final


@lru_cache(maxsize=None)
def _get_one_legal_entity(c, print_err_msg=True):
    if len(c) == 2:
        n = "n2"
    elif len(c) == 4:
        n = "n3"
    else:
        raise ValueError(
            "!!! Legal entity code should have 2 or 4 charaters !!!"
        )

    INSEE_legal_entity_n3 = (
        "https://api.insee.fr/metadonnees/codes/cj/" + n + "/"
    )

    with PynseeAPISession() as session:
        request = session.request_insee(
            api_url=INSEE_legal_entity_n3 + c,
            file_format="application/json;charset=utf-8",
            print_msg=print_err_msg,
        )

    data_request = request.json()

    data = pd.DataFrame(data_request, index=[0])

    return data



# -*- coding: utf-8 -*-
# Copyright : INSEE, 2021

import pandas as pd
from tqdm import trange
from pynsee.utils._make_dataframe_from_dict import _make_dataframe_from_dict


def _make_dataframe(data_request, main_key, query_number):

    list_dataframe = []

    try:
        data = data_request[main_key]
    except Exception:
        main_key_list = [
            key for key in list(data_request.keys()) if key != "header"
        ]
        main_key = main_key_list[0]
        data = data_request[main_key]

    for i in trange(len(data), desc="{} - Getting data".format(query_number)):
        idata = data[i]

        data_final = _make_dataframe_from_dict(idata)

        list_dataframe.append(data_final)

    data_final = pd.concat(list_dataframe)

    return data_final



from .get_sirene_data import get_sirene_data
from .search_sirene import search_sirene
from .get_dimension_list import get_dimension_list
from .get_sirene_relatives import get_sirene_relatives
from .sirenedataframe import SireneDataFrame

__all__ = [
    "get_sirene_data",
    "search_sirene",
    "get_dimension_list",
    "SireneDataFrame",
    "get_sirene_relatives",
]



# -*- coding: utf-8 -*-
# Copyright : INSEE, 2021

from functools import lru_cache
import pandas as pd

from pynsee.utils._paste import _paste
from pynsee.utils.requests_session import PynseeAPISession
from pynsee.localdata.get_area_list import get_area_list


@lru_cache(maxsize=None)
def _get_insee_one_area(area_type, codearea):

    df_list = get_area_list()
    list_available_codeareas = df_list.CODE.to_list()

    list_ZE20 = ["ZE2020", "zonesDEmploi2020", "ZoneDEmploi2020"]
    list_AAV20 = [
        "AAV2020",
        "airesDAttractionDesVilles2020",
        "AireDAttractionDesVilles2020",
    ]
    list_UU20 = ["UU2020", "unitesUrbaines2020", "UniteUrbaine2020"]

    list_type2 = [
        "zoneDEmploi2020",
        "aireDAttractionDesVilles2020",
        "uniteUrbaine2020",
    ]

    list_ZE20 = [s.lower() for s in list_ZE20]
    list_AAV20 = [s.lower() for s in list_AAV20]
    list_UU20 = [s.lower() for s in list_UU20]
    area_type = area_type.lower()

    type2 = ""
    if area_type in list_ZE20:
        type2 = "zoneDEmploi2020"
    if area_type in list_AAV20:
        type2 = "aireDAttractionDesVilles2020"
    if area_type in list_UU20:
        type2 = "uniteUrbaine2020"

    if type2 not in list_type2:
        geo_string = _paste(list_type2, collapse=" ")
        msg = "!!! Please choose area_type among:\n{}".format(geo_string)
        raise ValueError(msg)

    if codearea in list_available_codeareas:
        api_url = "https://api.insee.fr/metadonnees/geo/"
        api_url = api_url + type2 + "/" + codearea + "/descendants"
        with PynseeAPISession() as session:
            request = session.request_insee(
                api_url=api_url, file_format="application/json"
            )

        data = request.json()
        list_data_area = []
        for i in range(len(data)):
            df = pd.DataFrame(data[i], index=[0])
            list_data_area.append(df)

        data_area = pd.concat(list_data_area).reset_index(drop=True)

        ref_area_label = df_list.loc[df_list.CODE == codearea].TITLE
        ref_area_label = ref_area_label.reset_index(drop=True)[0]

        data_area = data_area.assign(
            ref_area_code=codearea, ref_area_label=ref_area_label
        )

        return data_area
    else:
        raise ValueError(
            "{} is not available in get_area_list('{}')".format(
                codearea, area_type
            )
        )



# -*- coding: utf-8 -*-

from typing import Any, Optional

from geopandas import GeoDataFrame

from .geofrdataframe import GeoFrDataFrame
from ._get_geodata import _get_geodata


def get_geodata(
    dataset_id: str,
    update: bool = False,
    crs: Any = "EPSG:3857",
    constrain_area: Optional[GeoDataFrame] = None,
) -> GeoFrDataFrame:
    """Get geographical data with identifier and from IGN API

    Args:
        id (str): data identifier from get_geodata_list function

        update (bool, optional): data is saved locally, set update=True to trigger an update. Defaults to False.

        crs (any valid :class:`~pyproj.crs.CRS` input, optional): CRS used for the geodata output. Defaults to 'EPSG:3857'.

        constrain_area (:class:`~geopandas.GeoDataFrame`, optional): GeoDataFrame used to constrain the area of interest. Defaults to None.

    .. versionchanged: 0.2.0

        Changed `polygon` and `crsPolygon` into a `constrain_area` :class:`~geopandas.GeoDataFrame`.

    Examples:
        >>> from pynsee.geodata import get_geodata_list, get_geodata
        >>> #
        >>> # Get a list of geographical limits of French administrative areas from IGN API
        >>> geodata_list = get_geodata_list()
        >>> #
        >>> # Get geographical limits of departments
        >>> df = get_geodata('ADMINEXPRESS-COG-CARTO.LATEST:departement')

    """
    polygon = None
    crsPolygon = "EPSG:4326"

    if constrain_area is not None:
        if constrain_area.crs.to_string() not in ("EPSG:3857", "EPSG:4326"):
            constrain_area = constrain_area.to_crs("EPSG:4326")

        polygon = constrain_area.union_all()
        crsPolygon = constrain_area.crs.to_string()

    return _get_geodata(
        dataset_id=dataset_id,
        update=update,
        crs=crs,
        ignore_error=False,
        polygon=polygon,
        crs_polygon=crsPolygon,
    )



# -*- coding: utf-8 -*-
# Copyright : INSEE, 2021

import pandas as pd


def _street_metadata():

    df = pd.DataFrame(
        {
            "typeVoieEtablissement": [
                "AIRE",
                "ALL",
                "AV",
                "BASE",
                "BD",
                "CAMI",
                "CAR",
                "CHE",
                "CHEM",
                "CHS",
                "CITE",
                "CLOS",
                "COIN",
                "COR",
                "COTE",
                "COUR",
                "CRS",
                "DOM",
                "DSC",
                "ECA",
                "ESP",
                "FG",
                "GARE",
                "GR",
                "HAM",
                "HLE",
                "ILOT",
                "IMP",
                "LD",
                "LOT",
                "MAR",
                "MTE",
                "PARC",
                "PAS",
                "PL",
                "PLAN",
                "PLN",
                "PLT",
                "PONT",
                "PORT",
                "PRO",
                "PRV",
                "QUA",
                "QUAI",
                "RES",
                "RLE",
                "ROC",
                "RPT",
                "RTE",
                "RUE",
                "SENTE",
                "SQ",
                "TOUR",
                "TPL",
                "TRA",
                "VLA",
                "VLGE",
                "VOIE",
                "ZA",
                "ZAC",
                "ZAD",
                "ZI",
                "ZONE",
                " ",
            ],
            "typeVoieEtablissementLibelle": [
                "Aire",
                "Allée",
                "Avenue",
                "Base",
                "Boulevard",
                "Cami",
                "Carrefour",
                "Chemin",
                "Cheminement",
                "Chaussée",
                "Cité",
                "Clos",
                "Coin",
                "Corniche",
                "Cote",
                "Cour",
                "Cours",
                "Domaine",
                "Descente",
                "Ecart",
                "Esplanade",
                "Faubourg",
                "Gare",
                "Grande Rue",
                "Hameau",
                "Halle",
                "Ilot",
                "Impasse",
                "Lieu dit",
                "Lotissement",
                "Marché",
                "Montée",
                "Parc",
                "Passage",
                "Place",
                "Plan",
                "Plaine",
                "Plateau",
                "Pont",
                "Port",
                "Promenade",
                "Parvis",
                "Quartier",
                "Quai",
                "Résidence",
                "Ruelle",
                "Rocade",
                "Rond Point",
                "Route",
                "Rue",
                "Sentier",
                "Square",
                "Tour",
                "Terre-Plein",
                "Traverse",
                "Villa",
                "Village",
                "Voie",
                "Zone artisanale",
                "Zone d'aménagement concerté",
                "Zone d'aménagement différé",
                "Zone industrielle",
                "Zone",
                " ",
            ],
        }
    )

    return df



import difflib
import hashlib
import logging
import os
import re
import tempfile
import warnings

from pynsee.download._download_pb import _download_pb
from pynsee.download._import_options import _import_options
from pynsee.download._get_dict_data_source import _get_dict_data_source
from pynsee.download._check_url import _check_url


logger = logging.getLogger(__name__)


def _download_store_file(tempdir: tempfile.TemporaryDirectory, id: str):
    """Download requested file and return some metadata that will
    be used

    Arguments:
        tempdir {tempfile.TemporaryDirectory} -- The temporary folder where the dataset should be written
        id {str} -- The name of the dataset desired

    Raises:
        ValueError: When the desired dataset
        is not found on insee.fr,
        an error is raised

    Returns:
        dict -- If everything works well, returns a dictionary
    """

    dict_data_source = _get_dict_data_source()
    caract = None
    if id in dict_data_source.keys():
        caract = dict_data_source[id]
    elif id.rsplit("_", 1)[-1].lower() == "latest":
        # allow to query a dataset with a custom "latest" TAG (for instance
        # "TAG_COM_..." only exists as a vintaged dataset) for stability
        # purposes
        root = id.rsplit("_", 1)[0]
        suggestions = {
            x
            for x in dict_data_source.keys()
            if re.match(root + ".?[0-9]{4}", x)
        }
        if suggestions:
            new_id = sorted(suggestions)[-1]
            logger.warning(
                "File %s not found. Switching to %s instead", id, new_id
            )
            caract = dict_data_source[new_id]

    if not caract:
        suggestions = difflib.get_close_matches(id, dict_data_source.keys())
        if len(suggestions) == 0:
            error_message = (
                "No file id found. Check metadata from get_file_list function"
            )
        else:
            error_message = f"Data name might be mispelled, \
                potential values are: {suggestions}"

        raise ValueError(error_message)

    filename = os.path.join(tempdir, "tempfile")

    url_found = _check_url(caract["lien"])

    _download_pb(url=url_found, fname=filename, total=caract["size"])

    # CHECKSUM MD5 ------------------------------------------

    with open(filename, "rb") as f:
        if hashlib.md5(f.read()).hexdigest() != caract["md5"]:
            warnings.warn(
                "File in insee.fr modified or corrupted during download"
            )

    # PREPARE PANDAS IMPORT ARGUMENTS -----------------------

    pandas_read_options = _import_options(tempdir, caract, filename)

    return {"result": caract, **pandas_read_options}



# -*- coding: utf-8 -*-
# Copyright : INSEE, 2021

import tempfile
from functools import lru_cache


@lru_cache(maxsize=None)
def _get_temp_dir():

    dirpath = tempfile.mkdtemp()
    return dirpath



# -*- coding: utf-8 -*-

from concurrent.futures import ThreadPoolExecutor, as_completed
import logging
import multiprocessing
import warnings
from typing import Any, Optional, Union
from xml.etree import ElementTree

import requests
from tqdm import tqdm
from pyproj.crs import CRS
from shapely import MultiPolygon, Polygon

from pynsee.utils.save_df import save_df
from pynsee.utils.requests_session import PynseeAPISession

from .geofrdataframe import GeoFrDataFrame


logger = logging.getLogger(__name__)


@save_df(cls=GeoFrDataFrame, day_lapse_max=90)
def _get_geodata(
    dataset_id: str,
    polygon: Optional[Union[MultiPolygon, Polygon]] = None,
    update: bool = False,
    silent: bool = False,
    crs: Any = "EPSG:3857",
    crs_polygon: Any = "EPSG:4326",
    ignore_error: bool = True,
) -> GeoFrDataFrame:
    """
    Get geographical data with identifier and from IGN API

    Args:
        id (str): _description_
        polygon (Polygon, optional): Polygon used to constrain the area of interest. Defaults to None.
        update (bool, optional): data is saved locally, set update=True to trigger an update. Defaults to False.
        silent (bool, optional): whether to print warnings or not. Defaults to False.
        crs (any valid :class:`~pyproj.crs.CRS` entry, optional): CRS used for the geodata output. Defaults to 'EPSG:3857'.
        crs_polygon (any valid :class:`~pyproj.crs.CRS` entry, optional): CRS used for `polygon`. Defaults to 'EPSG:4326'.
        ignore_error (boo, optional): whether to ignore errors and return an empty GeoDataFrame. Defaults to True.

    Examples:
        >>> from pynsee.geodata import get_geodata_list, get_geodata
        >>> #
        >>> # Get a list of geographical limits of French administrative areas from IGN API
        >>> geodata_list = get_geodata_list()
        >>> #
        >>> # Get geographical limits of departments
        >>> df = get_geodata('ADMINEXPRESS-COG-CARTO.LATEST:departement')

    Returns: GeoFrDataFrame
    """
    crs = CRS.from_user_input(crs).to_string()
    crs_polygon = CRS.from_user_input(crs_polygon).to_string()

    if crs_polygon not in ("EPSG:3857", "EPSG:4326"):
        raise ValueError(
            "crs_polygon must be either 'EPSG:3857' or 'EPSG:4326'"
        )

    bbox = ""

    # add bounding box to link if polygon provided
    if polygon is not None:
        bounds = polygon.bounds
        bounds = [str(b) for b in bounds]

        if crs_polygon == "EPSG:3857":
            bounds = [
                bounds[0],
                bounds[1],
                bounds[2],
                bounds[3],
                "urn:ogc:def:crs:" + crs_polygon,
            ]
        else:
            bounds = [
                bounds[1],
                bounds[0],
                bounds[3],
                bounds[2],
                "urn:ogc:def:crs:" + crs_polygon,
            ]

        bbox = f"&BBOX={','.join(bounds)}"

    geoportail = "https://data.geopf.fr"
    fmt = "application/json"

    urlhits = (
        "{url}/wfs/ows?service=WFS&version=2.0.0"
        "&request=GetFeature&typenames={typename}&resultType=hits" + bbox
    )

    urldata = (
        "{url}/wfs/?service=WFS&version=2.0.0"
        "&request=GetFeature&typenames={typename}"
        "&outputFormat={fmt}&startIndex={start}&count={count}"
        "&srsName={crs}" + bbox
    )

    hits = urlhits.format(url=geoportail, typename=dataset_id)

    with PynseeAPISession() as session:
        try:
            data = session.get(hits, verify=False)
        except requests.exceptions.RequestException as exc:
            if ignore_error:
                message = "Request failed"
                res = exc.response

                if hasattr(res, "status_code"):
                    message += f" (error {res.status_code}): {res.reason}"
                else:
                    message += str(exc)

                message += f"\nFaulty URL: {hits}."

                warnings.warn(message, category=RuntimeWarning, stacklevel=2)

                return GeoFrDataFrame()

            raise exc

        root = ElementTree.fromstring(data.content)
        num_hits = int(root.attrib["numberMatched"])

        data = []
        count = 5000  # max items returned by https://data.geopf.fr/wfs

        # iterate if necessary
        if num_hits > count:
            num_calls = num_hits // count

            if num_calls * count < num_hits:
                num_calls += 1

            n_proc = min(num_calls, 7, multiprocessing.cpu_count())

            link = urldata.format(
                url=geoportail,
                typename=dataset_id,
                fmt=fmt,
                start="{start}",
                count="{count}",
                crs=crs,
            )

            args = ((link, session, count, i) for i in range(num_calls))

            with ThreadPoolExecutor(max_workers=n_proc) as pool:

                submitted = [pool.submit(_make_request, x) for x in args]
                list_req = [
                    future.result()
                    for future in tqdm(
                        as_completed(submitted), total=num_calls
                    )
                ]

            for r, link in list_req:
                _check_request_update_data(data, r, link, polygon)
        else:
            link = urldata.format(
                url=geoportail,
                typename=dataset_id,
                fmt=fmt,
                start=0,
                count=num_hits,
                crs=crs,
            )

            r = session.get(link, verify=False)

            _check_request_update_data(data, r, link, polygon)

    if not data:
        warnings.warn(
            "Error 200: Valid request returned empty result",
            category=RuntimeWarning,
            stacklevel=2,
        )

        return GeoFrDataFrame()

    gdf = GeoFrDataFrame.from_features(data, crs=crs)

    # drop data outside polygon
    if polygon is not None:
        if gdf.crs != crs_polygon:
            gdf = gdf.to_crs(crs_polygon)

        return (
            gdf.iloc[gdf.sindex.query(polygon, predicate="intersects")]
            .to_crs(crs)
            .reset_index(drop=True)
        )

    return gdf


def _make_request(
    arg: tuple[str, requests.Session, int, int, int]
) -> tuple[requests.Response, str]:
    """Make the request inside the multiprocessing.Pool"""
    urldata, session, count, i = arg

    start = i * count
    link = urldata.format(start=start, count=count)

    return session.get(link, verify=False), link


def _check_request_update_data(
    data: list,
    r: requests.Request,
    link: str,
    polygon: Optional[Union[MultiPolygon, Polygon]],
) -> None:
    """
    Check that the request succeeded and update `data` inplace.
    """
    if not r.ok:
        raise requests.RequestException(
            f"The following query raise an error {r.status_code}: {link}"
        )

    json = r.json()["features"]

    if json:
        data.extend(json)
    else:
        msg = f"Query is correct but no data found: {link}"

        if polygon is not None:
            msg += (
                "\nCheck that crs_polygon argument corresponds "
                "to polygon data!"
            )

        logger.error(msg)



import pandas as pd
import re

from pynsee.utils.requests_session import PynseeAPISession
from pynsee.utils._make_dataframe_from_dict import _make_dataframe_from_dict
from pynsee.utils.HiddenPrints import HiddenPrints
from .sirenedataframe import SireneDataFrame


def get_sirene_relatives(*siret):
    """Find parent or child entities for one siret entity (etablissement)

    Args:
        siret (str or list): siret or list of siret codes

    Raises:
        ValueError: siret should be str or list

    Returns:
        pandas.DataFrame: dataframe containing the query content

    Examples:
        >>> # find parent or child entities for one siret entity (etablissement)
        >>> data = get_sirene_relatives('00555008200027')
        >>> data = get_sirene_relatives(['39860733300059', '00555008200027'])
    """

    list_siret = []

    for id in range(len(siret)):
        if isinstance(siret[id], list):
            list_siret += siret[id]
        elif isinstance(siret[id], pd.core.series.Series):
            list_siret += siret[id].to_list()
        elif isinstance(siret[id], str):
            list_siret += [siret[id]]
        else:
            list_siret += [str(siret[id])]

    types = ["siretEtablissementPredecesseur", "siretEtablissementSuccesseur"]
    list_df = []

    for s in range(len(list_siret)):
        for i in range(len(types)):

            criteria = types[i] + ":" + re.sub(r"\s+", "", list_siret[s])
            query = (
                "https://api.insee.fr/api-sirene/3.11/siret/liensSuccession"
                f"?q={criteria}"
            )
            try:
                with HiddenPrints():
                    with PynseeAPISession() as session:
                        result = session.request_insee(
                            api_url=query,
                            file_format="application/json;charset=utf-8",
                            raise_if_not_ok=True,
                            print_msg=False,
                        )

                    json = result.json()
            except Exception:
                pass
            else:
                list_df += [_make_dataframe_from_dict(json)]

    if len(list_df) > 0:
        df = SireneDataFrame(pd.concat(list_df).reset_index(drop=True))

        for c in ["statut", "message", "nombre", "total", "debut"]:
            if c in df.columns:
                del df[c]

        return df

    raise ValueError(
        "Neither parent nor child entities were found for any entity"
    )



import logging
import math
import warnings
from typing import Optional

import pandas as pd
from geopandas import GeoDataFrame
from shapely.affinity import rotate
from shapely.geometry import Point, LineString

from ._make_offshore_points import _make_offshore_points
from ._rescale_geom import _rescale_geom
from ._get_center import _get_center
from ._add_insee_dep import _add_insee_dep


logger = logging.getLogger(__name__)


def transform_overseas(
    gdf: GeoDataFrame,
    departement: tuple[str, ...] = ("971", "972", "974", "973", "976"),
    factor: tuple[Optional[float], ...] = (None, None, None, 0.35, None),
    center: tuple[float, float] = (-133583.39, 5971815.98),
    radius: float = 650000,
    angle: float = 1 / 9 * math.pi,
    startAngle: float = math.pi * (1 - 1.5 * 1 / 9),
) -> GeoDataFrame:
    """
    Move overseas departements closer to metropolitan France

    Args:
        departement (tuple, optional): list of departements to be moved, overseas departement list is used by default

        factor (tuple, optional): make departements bigger or smaller, it should correspond to the departement list.
        This parameter is used by shapely.affinity.scale function, please refer to its documentation to choose the value.
        By default, only Guyane's size is reduced. If the value is None, no rescaling is performed.

        center (tuple, optional): center point from which offshore points are computed to move overseas departement
        It should be defined as a (longitude, latitude) point in crs EPSG:3857

        radius (float, optional): radius used with center point to make offshore points, distance in meter

        angle (float, optional): angle used between offshore points, by default it is pi/9

        startAngle (float, optional): start angle defining offshore points, by default it is pi * (1 - 1.5 * 1/9))

    Notes:
        by default translate method focuses on overseas departement, but it can be used to move
        any departement anywhere on the map

    Examples:
        >>> from pynsee.geodata import get_geodata_list, get_geodata
        >>> #
        >>> # Get a list of geographical limits of French administrative areas from IGN API
        >>> geodata_list = get_geodata_list()
        >>> #
        >>> # Get geographical limits of departments
        >>> gdf = get_geodata('ADMINEXPRESS-COG-CARTO.LATEST:departement')
        >>> #
        >>> # Move overseas departements closer to metropolitan France
        >>> dfTranslate = gdf.translate()
    """
    init_crs = gdf.crs

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        if init_crs != "EPSG:3857":
            logging.warning("Converting GeoDataFrame to EPSG:3857.")
            gdf = gdf.to_crs("EPSG:3857")

        geocol = "insee_dep_geometry"

        if "id" in gdf and gdf["id"].str.match("DEPARTEM").all():
            # this is already the departments, work on geometry
            geocol = gdf.geometry.name
        else:
            gdf = _add_insee_dep(gdf.copy())

        offshore_points = _make_offshore_points(
            center=Point(center),
            list_ovdep=departement,
            radius=radius,
            angle=angle,
            startAngle=startAngle,
        )

        # keep only specific departements
        list_new_dep = [gdf.loc[~gdf.insee_dep.isin(departement)]]

        for i, d in enumerate(departement):
            ovdep = gdf.loc[gdf.insee_dep.values == d].reset_index(drop=True)

            if ovdep.empty:
                logger.warning("%s is missing from insee_dep column !", d)
            else:
                # get the center of the department to rescale the geometry
                # properly
                center = _get_center(ovdep, geocol=geocol)

                if factor[i] is not None:
                    _rescale_geom(ovdep, factor=factor[i], center=center)

                center_x, center_y = center

                xoff = offshore_points[i].coords.xy[0][0] - center_x
                yoff = offshore_points[i].coords.xy[1][0] - center_y

                ovdep.loc[:, ovdep.geometry.name] = ovdep.geometry.translate(
                    xoff=xoff, yoff=yoff
                )

                list_new_dep.append(ovdep)

        gdf = pd.concat(list_new_dep, ignore_index=True)

        return gdf.drop(columns=["insee_dep_geometry"], errors="ignore")


def zoom(
    gdf: GeoDataFrame,
    departement: tuple[str, ...] = ("75", "92", "93", "94"),
    center: tuple[float, float] = (-133583.39, 5971815.98),
    radius: float = 650000,
    startAngle: float = math.pi * (1 - 2.5 * 1 / 9),
    factor: float = 2,
) -> GeoDataFrame:
    """Zoom on parisian departements

    Args:
        departement (list, optional): list of departements to be moved, departements closest to Paris are selected by default

        center (tuple, optional): center point from which an offshore point is computed to move Parisian departements
        It should be defined as a (longitude, latitude) point in crs EPSG:3857

        radius (float, optional): radius used with center point to make offshore point, distance in meter

        startAngle (float, optional): start angle defining offshore point, by default it is pi * (1 - 2.5 * 1/9))

        factor (float, optional): make departements bigger or smaller.
        This parameter is used by shapely.affinity.scale function, please refer to its documentation to choose the value.

    Notes:
        by default zoom method focuses on the closest departements to Paris, but the function can be used to make a zoom on
        any departement anywhere on the map

    Examples:
        >>> from pynsee.geodata import get_geodata_list, get_geodata
        >>> #
        >>> # Get a list of geographical limits of French administrative areas from IGN API
        >>> geodata_list = get_geodata_list()
        >>> #
        >>> # Get geographical limits of departments
        >>> df = get_geodata('ADMINEXPRESS-COG-CARTO.LATEST:departement')
        >>> #
        >>> # Zoom on parisian departements
        >>> dfZoom = df.zoom()
    """
    if all([x in gdf.columns for x in ["insee_dep", "geometry"]]):
        zoomDep = gdf[gdf["insee_dep"].isin(departement)].reset_index(
            drop=True
        )

        if len(zoomDep.index) > 0:
            _rescale_geom(zoomDep, factor=factor)
            end = Point(center[0] + radius, center[1])
            line = LineString([center, end])

            line = rotate(line, startAngle, origin=center, use_radians=True)
            endPoint = Point(line.coords[1])
            center = _get_center(zoomDep)

            xoff = endPoint.coords.xy[0][0] - center[0]
            yoff = endPoint.coords.xy[1][0] - center[1]

            zoomDep["geometry"] = zoomDep["geometry"].translate(
                xoff=xoff, yoff=yoff
            )

            gdf = pd.concat([gdf, zoomDep]).reset_index(drop=True)

    return gdf



# -*- coding: utf-8 -*-
# Copyright : INSEE, 2021

from functools import lru_cache
import pandas as pd

from pynsee.utils.requests_session import PynseeAPISession


@lru_cache(maxsize=None)
def _get_insee_local_onegeo(variables, dataset_version, nivgeo, codegeo):

    count_sep = variables.count("-")
    modalite = ".all" * (count_sep + 1)

    link = "https://api.insee.fr/donnees-locales/donnees/"
    link = link + "geo-{}@{}/{}-{}{}".format(
        variables, dataset_version, nivgeo, codegeo, modalite
    )

    with PynseeAPISession() as session:
        request = session.request_insee(
            api_url=link, file_format="application/json;charset=utf-8"
        )

    data_request = request.json()

    Cellule = data_request["Cellule"]
    Variable = data_request["Variable"]
    Croisement = data_request["Croisement"]

    dataset_version = Croisement["JeuDonnees"]["code"]
    dataset_name = Croisement["JeuDonnees"]["Source"]
    data_date = Croisement["JeuDonnees"]["Annee"]

    list_data = []

    for i in range(len(Cellule)):
        dico = {**Cellule[i]["Zone"], **Cellule[i]["Mesure"]}
        modalite = Cellule[i]["Modalite"]

        for m in range(len(modalite)):
            try:
                dico_added = {modalite[m]["variable"]: modalite[m]["code"]}
            except Exception:
                dico_added = {modalite["variable"]: modalite["code"]}
            dico = {**dico, **dico_added}

        dico["OBS_VALUE"] = Cellule[i]["Valeur"]
        dico = {k: v for k, v in dico.items() if len(v) != 0}
        try:
            df = pd.DataFrame(dico, index=[0])
        except Exception:
            df = pd.DataFrame(dico)

        list_data.append(df)

    data = pd.concat(list_data)

    for i in range(len(Variable)):

        list_dict_var = []
        values = Variable[i]["Modalite"]
        for d in range(len(values)):
            df_dict = pd.DataFrame(values[d], index=[0])
            list_dict_var.append(df_dict)

        var_name = Variable[i]["code"]

        df = (
            pd.concat(list_dict_var)
            .reset_index(drop=True)
            .drop(columns=["variable"])
            .rename(columns={"code": var_name})
        )

        data[f"{var_name}_label"] = Variable[i]["Libelle"]

        data = data.merge(df, on=var_name, how="left")
    data = data.assign(
        DATASET_VERSION=dataset_version,
        DATASET_NAME=dataset_name,
        DATA_DATE=data_date,
    )

    data.rename(
        columns={
            "codgeo": "CODEGEO",
            "nivgeo": "NIVGEO",
            "code": "UNIT",
            "value": "UNIT_label_fr",
        },
        inplace=True,
    )

    data["OBS_VALUE"] = pd.to_numeric(data["OBS_VALUE"])

    return data



# -*- coding: utf-8 -*-

from functools import lru_cache

from pynsee.geodata.get_geodata import get_geodata


@lru_cache(maxsize=None)
def get_population():
    """Get population data on all French communes (cities)

    Examples:
        >>> from pynsee.localdata import get_population
        >>> pop = get_population()
    """

    df = get_geodata("ADMINEXPRESS-COG-CARTO.LATEST:commune")

    return df



# -*- coding: utf-8 -*-
# Copyright : INSEE, 2021

from functools import lru_cache
from tqdm import trange
import pandas as pd
import numpy as np
import re

from pynsee.localdata._find_latest_local_dataset import (
    _find_latest_local_dataset,
)
from pynsee.localdata._get_insee_local_onegeo import _get_insee_local_onegeo
from pynsee.localdata.get_geo_list import get_geo_list
from pynsee.utils.save_df import save_df

import logging

logger = logging.getLogger(__name__)


@lru_cache(maxsize=None)
def _warning_nivgeo(nivgeo):
    if nivgeo == "DEP":
        nivgeo_label = "departements"
        logger.info(f"By default, the query is on all {nivgeo_label}")
    elif nivgeo == "REG":
        nivgeo_label = "regions"
        logger.info(f"By default, the query is on all {nivgeo_label}")
    elif nivgeo == "FE":
        logger.info("By default, the query is on all France territory")


@save_df(day_lapse_max=90)
def get_local_data(
    variables,
    dataset_version,
    nivgeo="FE",
    geocodes=["1"],
    update=False,
    silent=False,
    backwardperiod=6,
):
    """Get INSEE local numeric data

    Args:
        variables (str): one or several variables separated by an hyphen (see get_local_metadata)

        dataset_version (str): code of a dataset version (see get_local_metadata), if dates are replaced by 'latest' the function triggers a loop to find the latest data available (examples: 'GEOlatestRPlatest', 'GEOlatestFLORESlatest').

        nivgeo (str): code of kind of French administrative area (see get_nivgeo_list), by default it is 'FE' ie all France

        geocodes (list): code one specific area (see get_geo_list), by default it is ['1'] ie all France

        update (bool): data is saved locally, set update=True to trigger an update

        silent (bool, optional): Set to True to disable messages printed in log info

        backwardperiod (int, optional): this arg is used only whenever the latest data is searched, it specifies the number of past years the loop should run through.

    Raises:
        ValueError: Error if geocodes is not a list

    Examples:
        >>> from pynsee.localdata import get_local_metadata, get_nivgeo_list, get_geo_list, get_local_data
        >>> metadata = get_local_metadata()
        >>> nivgeo = get_nivgeo_list()
        >>> departement = get_geo_list('departements')
        >>> #
        >>> data_all_france = get_local_data(dataset_version='GEO2020RP2017',
        >>>                        variables =  'SEXE-DIPL_19')
        >>> #
        >>> data_91_92 = get_local_data(dataset_version='GEO2020RP2017',
        >>>                        variables =  'SEXE-DIPL_19',
        >>>                        nivgeo = 'DEP',
        >>>                        geocodes = ['91','92'])
        >>> #
        >>> # get latest data for from RP (Recensement / Census) on socio-professional categories by sexe in Paris
        >>> data_paris = get_local_data(dataset_version='GEOlatestRPlatest',
        >>>                        variables =  'CS1_8-SEXE',
        >>>                        nivgeo = 'COM',
        >>>                        geocodes = '75056')
    """

    if isinstance(geocodes, pd.core.series.Series):
        geocodes = geocodes.to_list()

    if isinstance(geocodes, str):
        geocodes = [geocodes]

    if not isinstance(geocodes, list):
        raise ValueError("!!! geocodes must be a list !!!")

    if (geocodes == ["1"]) or (geocodes == ["all"]) or (geocodes == "all"):
        if nivgeo == "DEP":
            departement = get_geo_list("departements")
            geocodes = departement.CODE.to_list()
            _warning_nivgeo(nivgeo)
        elif nivgeo == "REG":
            reg = get_geo_list("regions")
            geocodes = reg.CODE.to_list()
            _warning_nivgeo(nivgeo)
        elif nivgeo == "FE":
            _warning_nivgeo(_warning_nivgeo)
        elif nivgeo != "METRODOM":
            logger.warning("Please provide a list with geocodes argument !")

    #
    # LATEST AVAILABLE DATASET OPTION
    #

    pattern = re.compile(".*latest$")

    if pattern.match(dataset_version):

        dataset_version = _find_latest_local_dataset(
            dataset_version,
            variables,
            nivgeo,
            geocodes[0],
            update,
            backwardperiod,
        )

    list_data_all = []

    for cdg in trange(len(geocodes), desc="Getting data"):

        codegeo = geocodes[cdg]
        df_default = pd.DataFrame(
            {"CODEGEO": codegeo, "OBS_VALUE": np.nan}, index=[0]
        )

        try:
            df = _get_insee_local_onegeo(
                variables, dataset_version, nivgeo, codegeo
            )

        except Exception:
            df = df_default

        list_data_all.append(df)

    data_final = pd.concat(list_data_all).reset_index(drop=True)

    if data_final.equals(df_default):
        logger.error("Error or no data found !")

    return data_final



# -*- coding: utf-8 -*-

from functools import lru_cache
import pandas as pd
import xml.etree.ElementTree as ET
import difflib
import numpy as np

from pynsee.geodata._get_capabilities import _get_capabilities
from pynsee.utils._clean_str import _clean_str


@lru_cache(maxsize=None)
def _get_full_list_wfs(topic="", version="2.0.0"):

    raw_data_file = _get_capabilities(
        key=topic, version=version, service="wfs"
    )

    root = ET.parse(raw_data_file).getroot()

    list_var = ["FeatureTypeList"]

    find = False

    for i in range(len(root)):
        for var in list_var:
            if _clean_str(root[i].tag) == var:
                data = root[i]
                find = True
                break
        if find:
            break

    list_df = []

    for i in range(len(data)):

        df = data[i]
        d = {}

        list_var0 = ["Name", "Title", "DefaultCRS", "Abstract"]
        list_var1 = ["WGS84BoundingBox", "Keywords"]
        list_subvar = ["LowerCorner", "UpperCorner", "Keyword"]

        for j in range(len(df)):

            for var in list_var0:
                if _clean_str(df[j].tag) == var:
                    d[var] = df[j].text

            for var in list_var1:
                if _clean_str(df[j].tag) == var:
                    for z in range(len(df[j])):
                        for subvar in list_subvar:
                            if _clean_str(df[j][z].tag) == subvar:
                                d[subvar] = df[j][z].text

        df2 = pd.DataFrame(d, index=[0])

        list_df.append(df2)

    if len(list_df) > 0:
        data_all = (
            pd.concat(list_df).reset_index(drop=True).dropna(axis=0, how="all")
        )
    else:
        data_all = list_df

    if "Keyword" in data_all.columns:
        list_keyword = [
            k for k in list(data_all["Keyword"].unique()) if k is not np.nan
        ]

        string_match_list = difflib.get_close_matches(
            "unites administratives", list_keyword, n=1
        )

        if len(string_match_list) > 0:

            string_match = string_match_list[0]  # NOQA

            data_all = data_all.query("Keyword == @string_match").reset_index(
                drop=True
            )

    return data_all



# -*- coding: utf-8 -*-
# Copyright : INSEE, 2021

from functools import lru_cache
import io
import pandas as pd
import xml.etree.ElementTree as ET

from pynsee.utils._paste import _paste
from pynsee.utils.requests_session import PynseeAPISession


@lru_cache(maxsize=None)
def _get_geo_relation(geo, code, relation, date=None, type=None):

    # relation_list = ['ascendants', 'descendants', 'suivants', 'precedents', 'projetes']
    # list_available_geo = ['communes', 'regions', 'departements',
    #                       'arrondissements', 'arrondissementsMunicipaux']
    # code = '11'
    # geo = 'region'
    # relation = 'descendants'

    # idf = _get_geo_relation('region', "11", 'descendants')
    # essonne = _get_geo_relation('region', "11", 'ascendants')

    api_url = (
        "https://api.insee.fr/metadonnees/geo/"
        + geo
        + "/"
        + code
        + "/"
        + relation
    )

    parameters = ["date", "type"]

    list_addded_param = []
    for param in parameters:
        if eval(param) is not None:
            list_addded_param.append(param + "=" + str(eval(param)))

    added_param_string = ""
    if len(list_addded_param) > 0:
        added_param_string = "?" + _paste(list_addded_param, collapse="&")
        api_url = api_url + added_param_string

    with PynseeAPISession() as session:
        results = session.request_insee(api_url=api_url)

    raw_data_file = io.BytesIO(results.content)

    root = ET.parse(raw_data_file).getroot()

    n_geo = len(root)

    list_geo_relation = []

    for igeo in range(n_geo):
        n_var = len(root[igeo])

        dict_var = {}

        for ivar in range(n_var):
            dict_var[root[igeo][ivar].tag] = root[igeo][ivar].text

        dict_var = {**dict_var, **root[igeo].attrib}
        df_relation = pd.DataFrame(dict_var, index=[0])
        list_geo_relation.append(df_relation)

    df_relation_all = pd.concat(list_geo_relation)
    df_relation_all = df_relation_all.assign(geo_init=code)

    return df_relation_all



# -*- coding: utf-8 -*-

import pandas as pd
from shapely.geometry import shape
import warnings

import logging

logger = logging.getLogger(__name__)


def _geojson_parser(data):

    data_list = []
    list_shapes = []

    for c in range(len(data)):

        df = data[c]["properties"]

        df2 = pd.DataFrame(
            {
                f: df[f]
                for f in df.keys()
                if f not in (["geometry", "bbox"])
                and (type(df[f]) is not list)
            },
            index=[0],
        )

        for k in df.keys():
            dfk = df[k]
            if type(dfk) is list:
                try:
                    df2[k] = [df[k]]
                except Exception:
                    pass

        geom = data[c]["geometry"]["coordinates"]

        data_type = data[c]["geometry"]["type"]

        try:
            if "id" not in df2.columns:
                df2["id"] = data[c]["id"]
        except Exception:
            pass

        Shape = shape({"type": data_type, "coordinates": geom})
        # list_shapes = [Shape.geoms[x] for x in range(len(Shape.geoms))]

        if data_type in [
            "MultiLineString",
            "MultiPolygon",
            "MultiPoint",
            "LineString",
            "Polygon",
            "Point",
        ]:
            list_shapes += [Shape]
            data_list.append(df2)
        else:
            logger.warning(f"Unsupported shape {data_type} has been removed")

    data_all = pd.concat(data_list)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        data_all["geometry"] = list_shapes

    return data_all



# -*- coding: utf-8 -*-
# Copyright : INSEE, 2021

import json
import logging
import os
from typing import Optional

import requests

from pynsee.utils.requests_session import PynseeAPISession
from pynsee.constants import (
    CONFIG_FILE,
    SIRENE_KEY,
    HTTPS_PROXY_KEY,
    HTTP_PROXY_KEY,
)


logger = logging.getLogger(__name__)


def opener(path, flags):
    return os.open(path, flags, 0o600)


def init_conn(
    sirene_key: Optional[str] = None,
    http_proxy: Optional[str] = None,
    https_proxy: Optional[str] = None,
) -> None:
    """Save your credentials to connect to INSEE APIs, subscribe to api.insee.fr

    Args:
        sirene_key (str, optional): user's key for sirene API
        http_proxy (str, optional): Proxy server address, e.g. 'http://my_proxy_server:port'. Defaults to "".
        https_proxy (str, optional): Proxy server address, e.g. 'http://my_proxy_server:port'. Defaults to "".

    Notes:
        Environment variables can be used instead of init_conn function

    Examples:
        >>> from pynsee.utils import init_conn
        >>> init_conn(sirene_key="my_sirene_key")
        >>> #
        >>> # if the user has to use a proxy server use http_proxy and https_proxy arguments as follows:
        >>> from pynsee.utils import init_conn
        >>> init_conn(sirene_key="my_sirene_key",
        >>>           http_proxy="http://my_proxy_server:port",
        >>>           https_proxy="http://my_proxy_server:port")
        >>> #
        >>> # Alternativety you can use directly environment variables as follows:
        >>> # Beware not to commit your credentials!
        >>> import os
        >>> os.environ['sirene_key'] = 'my_sirene_key'
        >>> os.environ['http_proxy'] = "http://my_proxy_server:port"
        >>> os.environ['https_proxy'] = "http://my_proxy_server:port"
        >>> init_conn()
    """
    logger.debug("SHOULD GET LOGGING")

    try:
        with open(CONFIG_FILE, opener=opener, encoding="utf8") as f:
            init_config = json.load(f)
    except FileNotFoundError:
        init_config = {}

    with PynseeAPISession(
        sirene_key=sirene_key, http_proxy=http_proxy, https_proxy=https_proxy
    ) as session:
        try:
            invalid_requests = session._test_connections()
        except (ValueError, requests.exceptions.RequestException):
            try:
                os.remove(CONFIG_FILE)
            except FileNotFoundError:
                pass
            raise

    if invalid_requests:
        logger.error(
            "Invalid credentials, the following APIs returned error codes, "
            "please make sure you subscribed to them (if you need those):\n"
            f"{invalid_requests}"
        )
    else:
        logger.info(
            "Subscription to all INSEE's APIs has been successfull\n"
            "Unless the user wants to change the key or secret, using this "
            "function is no longer needed as the credentials will be saved "
            f"locally here:\n{CONFIG_FILE}"
        )

    if invalid_requests and ("Sirene" in invalid_requests):

        if not sirene_key:
            # user has probably no use for SIRENE API and has already
            # been warned about this
            sirene_key = None
        else:
            # invalid sirene key, restore from previous valid stored key
            try:
                sirene_key = init_config[SIRENE_KEY]
            except KeyError:
                sirene_key = None
            else:
                logger.warning("Previous SIRENE key has been restored.")

    config = {
        SIRENE_KEY: sirene_key,
        HTTP_PROXY_KEY: http_proxy,
        HTTPS_PROXY_KEY: https_proxy,
    }

    if {"Sirene"}.issuperset(invalid_requests):
        with open(CONFIG_FILE, "w", opener=opener, encoding="utf8") as f:
            json.dump(config, f)
        logger.info("Credentials have been saved.")



# -*- coding: utf-8 -*-
# Copyright : INSEE, 2021

from pynsee.utils.requests_session import PynseeAPISession
from tqdm import trange

import re
import pandas as pd


def get_definition(ids):
    """Get the definition of a concept from its identifier

    Args:
        ids (list): a list of concept identifiers

    Raises:
        ValueError: an error is raised if ids is not a list

    Examples:
        >>> from pynsee.metadata import get_definition_list, get_definition
        >>> def_list = get_definition_list()
        >>> # geographic areas definition
        >>> geo_definitions = get_definition(['c1468', 'c1282', 'c1762',
        >>>                       'c1501', 'c1346', 'c1502', 'c1912',
        >>>                       'c1361', 'c2173', 'c2070'])

    """
    if isinstance(ids, str):
        ids = [ids]

    if not isinstance(ids, list):
        raise ValueError("!!! ids must be a list or a str!!!")

    # ids = ['c1020', 'c1601']

    link = "https://api.insee.fr/metadonnees/concepts/definition"

    list_data = []

    def clean_definition(string):
        m = re.search("\\<p\\>.*\\<\\/p\\>", string)
        if m:
            string_selected = m.group(0)
            string_cleaned = (
                string_selected.replace("<p>", "")
                .replace("</p>", "")
                .replace("\xa0", " ")
            )
        else:
            string_cleaned = string
        return string_cleaned

    def extract_data(data, item1, i, item2):
        try:
            val = data[item1][i][item2]
        except Exception:
            val = None
        return val

    for i in trange(len(ids), desc="Getting data"):

        id = ids[i]
        query = link + "/" + id

        with PynseeAPISession() as session:
            request = session.request_insee(
                api_url=query, file_format="application/json"
            )

        try:
            data = request.json()

            title_fr = None
            title_en = None

            if data["intitule"][0]["langue"] == "fr":
                title_fr = extract_data(
                    data, item1="intitule", i=0, item2="contenu"
                )
                title_en = extract_data(
                    data, item1="intitule", i=1, item2="contenu"
                )

            def_fr = None
            def_en = None

            if data["definition"][0]["langue"] == "fr":
                def_fr = extract_data(
                    data, item1="definition", i=0, item2="contenu"
                )
                def_en = extract_data(
                    data, item1="definition", i=1, item2="contenu"
                )
                def_fr = clean_definition(def_fr)
                def_en = clean_definition(def_en)

            def_short_fr = None
            def_short_en = None

            try:
                if data["definitionCourte"][0]["langue"] == "fr":
                    def_short_fr = extract_data(
                        data, item1="definitionCourte", i=0, item2="contenu"
                    )
                    def_short_en = extract_data(
                        data, item1="definitionCourte", i=1, item2="contenu"
                    )
                    def_short_fr = clean_definition(def_short_fr)
                    def_short_en = clean_definition(def_short_en)
            except Exception:
                pass

            update = data["dateMiseAJour"]

            uri = data["uri"]

            df = pd.DataFrame(
                {
                    "ID": id,
                    "TITLE_FR": title_fr,
                    "TITLE_EN": title_en,
                    "DEFINITION_SHORT_FR": def_short_fr,
                    "DEFINITION_SHORT_EN": def_short_en,
                    "DEFINITION_FR": def_fr,
                    "DEFINITION_EN": def_en,
                    "UPDATE": update,
                    "URI": uri,
                },
                index=[0],
            )
        except Exception:
            df = pd.DataFrame(
                {
                    "ID": id,
                    "TITLE_FR": None,
                    "TITLE_EN": None,
                    "DEFINITION_SHORT_FR": None,
                    "DEFINITION_SHORT_EN": None,
                    "DEFINITION_FR": None,
                    "DEFINITION_EN": None,
                    "UPDATE": None,
                    "URI": None,
                },
                index=[0],
            )

        list_data.append(df)

    data_final = pd.concat(list_data)
    data_final = data_final.reset_index(drop=True)

    return data_final



import os
import sys

# source and credits:
# https://stackoverflow.com/a/45669280


class HiddenPrints:
    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, "w")

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout



# -*- coding: utf-8 -*-
# Copyright : INSEE, 2021

import pandas as pd

from pynsee.localdata._warning_local_data import _warning_local_data


def get_nivgeo_list():
    """Get a list of geographic levels

    Examples
        >>> from pynsee.localdata import get_nivgeo_list
        >>> nivgeo_list = get_nivgeo_list()
    """
    _warning_local_data()

    dict_ng = {
        "NIVGEO": [
            "COM",
            "DEP",
            "REG",
            "METRODOM",
            "FE",
            "ARR",
            "EPCI",
            "AAV2020",
            "UU2020",
            "ZE2020",
            "AU2010",
            "UU2010",
            "ZE2010",
        ],
        "NIVGEO_label_fr": [
            "communes et arrondissements municipaux",
            "départements",
            "régions",
            "France métropolitaine",
            "France",
            "arrondissements",
            "intercommunalités",
            "aires d'attraction des villes 2020",
            "unités urbaines 2020",
            "zones d'emploi 2020",
            "aires urbaines 2010",
            "unités urbaines 2010",
            "zones d'emploi 2010",
        ],
        "NIVGEO_label_en": [
            "municipalities and municipal districts",
            "departments",
            "regions",
            "metropolitan France",
            "France",
            "districts",
            "intermunicipal authorities",
            "functional areas 2020",
            "urban unit 2020",
            "employment zone 2020",
            "urban area 2010",
            "urban unit 2010",
            "employment zone 2010",
        ],
    }

    nivgeo = pd.DataFrame(dict_ng)
    return nivgeo



from typing import Optional
from geopandas import GeoDataFrame, GeoSeries


def _get_center(
    gdf: GeoDataFrame, geocol: Optional[str] = None
) -> tuple[float, float]:
    """
    Get the center of the geometries in a GeoDataFrame.

    Note
    ----
    This function may be applied on a column that is *not* the
    default geometry (i.e. not a ``GeoSeries``).
    """
    geocol = geocol or gdf.geometry.name

    geoseries = (
        gdf.geometry
        if geocol == gdf.geometry.name
        else GeoSeries(gdf[geocol], crs=gdf.crs)
    )

    minxdf, maxydf, maxxdf, minydf = geoseries.total_bounds

    return (maxxdf + minxdf) / 2, (maxydf + minydf) / 2



# -*- coding: utf-8 -*-

import pandas as pd
import datetime
from functools import lru_cache

from pynsee.utils.requests_session import PynseeAPISession
from pynsee.utils.save_df import save_df

import logging

logger = logging.getLogger(__name__)


@lru_cache(maxsize=None)
def _warning_get_area_projection():
    logger.info(
        "dateProjection is None, by default it is supposed to be the current "
        "date"
    )


@save_df(day_lapse_max=90)
def get_area_projection(
    area: str,
    code: str,
    date: str,
    dateProjection: str = None,
    silent: bool = False,
):
    """
    Get data about the area (valid at given `date` datetime) projected
    at `dateProjection` datetime.

    Args:
        area (str): case insensitive, area type, any of (
            'arrondissement',
            'arrondissementMunicipal',
            'commune',
            'departement',
            'region'
            )

        code (str): city code

        date (str): date used to analyse the data, format : 'AAAA-MM-JJ'.

        dateProjection (str, optional): date used to project the area into,
            format : 'AAAA-MM-JJ'.  If dateProjection is None, by default it
            is supposed to be the current date (ie projection into today's
            value)

        silent (bool, optional): Set to True, to disable messages printed in log info

    Examples:
        >>> from pynsee.localdata import get_area_projection
        >>> df = get_area_projection(
                code='01039',
                date='2020-01-01',
                dateProjection='2023-04-01'
                )
    """

    areas = {
        "arrondissement": "arrondissement",
        "arrondissementmunicipal": "arrondissementMunicipal",
        "commune": "commune",
        "departement": "departement",
        "intercommunalite": "intercommunalite",
        "region": "region",
    }
    if area.lower() not in areas:
        msg = f"territory must be one of {areas} - found '{area}' instead"
        raise ValueError(msg)
    else:
        area = areas[area.lower()]

    INSEE_localdata_api_link = "https://api.insee.fr/metadonnees/geo/"

    api_link = f"{INSEE_localdata_api_link}{area}/{code}/projetes?date={date}"

    if not dateProjection:
        _warning_get_area_projection()
        dateProjection = datetime.date.today().strftime("%Y-%m-%d")

    api_link += f"&dateProjection={dateProjection}"

    try:
        with PynseeAPISession() as session:
            request = session.request_insee(
                api_url=api_link, file_format="application/json"
            )

        data = pd.DataFrame(request.json())

    except Exception:
        logger.error(
            f"No data found for projection of area {area} of code {code} "
            f"from date {date} to date {dateProjection}"
        )
        data = pd.DataFrame()

    return data



from datetime import date
import logging
import re

import requests

from pynsee.utils.requests_session import PynseeAPISession

logger = logging.getLogger(__name__)


def _check_url(url):

    with PynseeAPISession() as session:

        try:
            check = session.get(url, stream=True, verify=False)
        except requests.exceptions.RequestException:
            pass
        else:
            return url

        logger.error(
            f"File not found on insee.fr:\n{url} - please open an issue on:\n"
            "https://github.com/InseeFrLab/pynsee"
        )

        try:
            list_string_split = url.split("/")
            filename = list_string_split[-1]

            list_potential_dates = []

            def get_close_dates_list(start_year, timespan=10):
                start_year = int(start_year)
                start_year_short = int(str(start_year)[-2:])

                list_close_year = (
                    list(range(start_year, start_year + timespan))
                    + list(range(start_year, start_year - timespan, -1))
                    + list(
                        range(start_year_short, start_year_short + timespan)
                    )
                    + list(
                        range(
                            start_year_short, start_year_short - timespan, -1
                        )
                    )
                )

                list_close_year = [str(y) for y in list_close_year]

                return list_close_year

            datefile = re.findall(r"2\d{3}|\d{2}", filename)[0]

            list_potential_dates += get_close_dates_list(datefile)

            current_year = date.today().year

            list_potential_dates += get_close_dates_list(current_year)
            list_potential_dates = list(dict.fromkeys(list_potential_dates))

            for d in list_potential_dates:
                filename2 = filename.replace(str(datefile), str(d))
                url2 = "/".join(
                    list_string_split[: (len(list_string_split) - 1)]
                    + [filename2]
                )
                try:
                    results = session.get(url2, stream=True, verify=False)
                except requests.exceptions.RequestException:
                    pass
                else:
                    break

                if d == list_potential_dates[-1]:
                    logger.warning("No other similar files have been found")
                    url2 = url
        except Exception:
            logger.error(
                "Error raised while trying to find another similar file"
            )

        if "url2" in locals():
            if url != url2:
                logger.warning(
                    f"The following file has been used instead:\n{url2}"
                )

        else:
            url2 = url

        return url2



# -*- coding: utf-8 -*-
# Copyright : INSEE, 2021

import functools
import gc


def clear_all_cache():
    """Clear the cache of all functions

    Notes:
        If the credentials provided fail to get a token from api.insee.fr even after a double check,
        try to clear the cache as the output of the function retrieving the token is cached even
        it is an error.

    Examples:
        >>> from pynsee.utils import clear_all_cache
        >>> clear_all_cache()
    """
    gc.collect()
    wrappers = [
        a
        for a in gc.get_objects()
        if isinstance(a, functools._lru_cache_wrapper)
    ]

    for wrapper in wrappers:
        wrapper.cache_clear()



import math
import warnings
from typing import Optional

from shapely.affinity import rotate
from shapely.geometry import Point, LineString


def _make_offshore_points(
    center: Optional[Point] = None,
    radius: float = 8,
    angle: float = 1 / 17,
    startAngle: float = math.pi * (1 - 1.5 * 1 / 17),
    list_ovdep: tuple[str, ...] = ("974", "971", "972", "973", "976"),
) -> list[Point]:
    """Create the center points for French overseas departments"""
    center = center or Point(-1, 47.181903)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        list_points = []

        for i in range(len(list_ovdep)):

            angle_rotated = startAngle + angle * i
            end = Point(center.x + radius, center.y)
            line = LineString([center, end])
            line = rotate(line, angle_rotated, origin=center, use_radians=True)

            list_points.append(Point(line.coords[1]))

        return list_points



import pandas as pd
from pynsee.utils._move_col_after import _move_col_before


def _get_A5_activity_label():

    A5 = pd.DataFrame(
        {
            "A5": ["AZ", "BE", "FZ", "GU", "OQ"],
            "TITLE_A5_FR": [
                "Agriculture, sylviculture et pêche",
                "Industrie manufacturière, industries extractives et autres",
                "Construction",
                "Services principalement marchands",
                "Services principalement non marchands",
            ],
            "TITLE_A5_EN": [
                "Agriculture, forestry, and fisheries",
                "Mining, quarrying and manufacturing",
                "Construction",
                "Mainly market services",
                "Mainly non-market services",
            ],
        }
    )

    return A5


def _add_A5_activity(df):

    df2 = df.copy()
    df2["A5"] = df2["A10"].copy()

    string = "|".join(["GI", "JZ", "KZ", "LZ", "LI", "MN", "RU"])
    df2["A5"] = df2["A5"].replace(to_replace=string, value="GU", regex=True)

    df2 = _move_col_before(df2, "A5", "A10")

    return df2



# -*- coding: utf-8 -*-
# Copyright : INSEE, 2021

import pandas as pd

from pynsee.utils.requests_session import PynseeAPISession
from pynsee.utils._paste import _paste
from pynsee.utils.save_df import save_df


@save_df(day_lapse_max=90)
def get_area_list(
    area=None, date=None, update=False, silent=False
) -> pd.DataFrame:
    """Get an exhaustive list of administrative areas : communes, departments, and urban, employment or functional areas

    Args:
        area (str, optional): Defaults to None, then get all values

        date (str): date of validity (AAAA-MM-DD)

        update (bool): locally saved data is used by default. Trigger an update with update=True.

        silent (bool, optional): Set to True, to disable messages printed in log info

    Raises:
        ValueError: Error if area is not available

    Examples:
        >>> from pynsee.localdata import get_area_list
        >>> area_list = get_area_list()
        >>> #
        >>> # get list of all communes in France
        >>> reg = get_area_list(area='regions')
    """

    list_available_area = [
        "departements",
        "regions",
        "communes",
        "communesAssociees",
        "communesDeleguees",
        "arrondissementsMunicipaux",
        "arrondissements",
        "zonesDEmploi2020",
        "airesDAttractionDesVilles2020",
        "unitesUrbaines2020",
        "collectivitesDOutreMer",
    ]
    area_string = _paste(list_available_area, collapse=" ")

    list_ZE20 = ["ZE2020", "zonesDEmploi2020", "ZoneDEmploi2020"]
    list_AAV20 = [
        "AAV2020",
        "airesDAttractionDesVilles2020",
        "AireDAttractionDesVilles2020",
    ]
    list_UU20 = ["UU2020", "unitesUrbaines2020", "UniteUrbaine2020"]

    list_ZE20 = list_ZE20 + [s.lower() for s in list_ZE20]
    list_AAV20 = list_AAV20 + [s.lower() for s in list_AAV20]
    list_UU20 = list_UU20 + [s.lower() for s in list_UU20]

    if area is not None:
        if area in list_ZE20:
            area = "zonesDEmploi2020"
        if area in list_AAV20:
            area = "airesDAttractionDesVilles2020"
        if area in list_UU20:
            area = "unitesUrbaines2020"

        if area not in list_available_area + [
            x.lower() for x in list_available_area
        ]:
            msg = (
                f"!!! {area} is not available\n"
                f"Please choose area among:\n{area_string}"
            )
            raise ValueError(msg)

        list_available_area = [area]

    list_data = []

    with PynseeAPISession() as session:

        for a in list_available_area:
            api_url = "https://api.insee.fr/metadonnees/geo/" + a
            if date:
                api_url += f"?date={date}"

            request = session.request_insee(
                api_url=api_url, file_format="application/json"
            )

            data = request.json()
            list_data += data

    data_all = pd.DataFrame(list_data)

    data_all.rename(
        columns={
            "code": "CODE",
            "uri": "URI",
            "dateCreation": "DATE_CREATION",
            "intituleSansArticle": "TITLE_SHORT",
            "type": "AREA_TYPE",
            "typeArticle": "DETERMINER_TYPE",
            "intitule": "TITLE",
            "dateSuppression": "DATE_DELETION",
        },
        inplace=True,
    )

    return data_all



from tqdm import trange
from pandas.api.types import CategoricalDtype

from ._get_geodata_with_backup import _get_geodata_with_backup


def _add_insee_dep(gdf):
    gdf = gdf.reset_index(drop=True)

    # option 1 : get insee_dep from id_com colum
    gdf = _add_insee_dep_from_id_com(gdf)

    # option 2 : get insee_dep for regions
    if "insee_dep" not in gdf:
        gdf = _add_insee_dep_region(gdf)

        # option 3 : get insee_dep from get_geodata  and polygon intersection
        if "insee_dep" not in gdf:
            gdf = _add_insee_dep_from_geodata(gdf)

    if "insee_dep_geometry" not in gdf:
        raise ValueError("Could not get department geometries.")

    return gdf


def _add_insee_dep_from_id_com(gdf):
    # add insee_dep column
    if "insee_com" in gdf.columns:
        gdf.loc[:, "insee_dep"] = [
            v[:3] if v.startswith("97") else v[:2]
            for v in gdf.insee_com.values
        ]
    elif "id_com" in gdf.columns:
        try:
            dataset_id = "ADMINEXPRESS-COG-CARTO.LATEST:commune"
            com = _get_geodata_with_backup(dataset_id).to_crs("EPSG:3857")

            com = com[["id", "insee_dep"]]
            com = com.rename(columns={"id": "id_com"})
            gdf = gdf.merge(com, on="id_com", how="left")
        except Exception:
            return gdf

    # get departments and add the geometry
    try:
        dataset_id = "ADMINEXPRESS-COG-CARTO.LATEST:departement"
        dep = _get_geodata_with_backup(dataset_id).to_crs("EPSG:3857")

        dep = dep[["insee_dep", "geometry"]]
        dep = dep.rename(columns={"geometry": "insee_dep_geometry"})

        gdf = gdf.merge(dep, on="insee_dep", how="left")
    except Exception:
        pass

    return gdf


def _add_insee_dep_region(gdf):
    try:
        if "id" in gdf.columns and all(gdf.id.str.match("^REGION")):
            dataset_id = "ADMINEXPRESS-COG-CARTO.LATEST:departement"

            # get departments, keep only one for each region
            dep = (
                _get_geodata_with_backup(dataset_id)
                .drop_duplicates(subset="insee_reg", keep="first")
                .to_crs("EPSG:3857")
            )

            dep = dep[["insee_dep", "insee_reg", "geometry"]]
            dep = dep.rename(columns={"geometry": "insee_dep_geometry"})
            gdf = gdf.merge(dep, on="insee_reg", how="left")
    except Exception:
        pass

    return gdf


def _add_insee_dep_from_geodata(gdf):
    try:
        if "insee_dep_geometry" not in gdf.columns:
            gdf = gdf.reset_index(drop=True)

            list_dep = []
            list_dep_geo = []

            dataset_id = "ADMINEXPRESS-COG-CARTO.LATEST:departement"
            dep_list = _get_geodata_with_backup(dataset_id).to_crs("EPSG:3857")

            list_ovdep = ["971", "972", "974", "973", "976"]
            list_other_dep = [
                d for d in dep_list.insee_dep if d not in list_ovdep
            ]
            dep_order = list_ovdep + list_other_dep

            dep_list["insee_dep"] = dep_list["insee_dep"].astype(
                CategoricalDtype(categories=dep_order, ordered=True)
            )

            dep_list = dep_list.sort_values(["insee_dep"]).reset_index(
                drop=True
            )

            for i in trange(len(gdf.index), desc="Finding departement"):
                geo = gdf.loc[i, "geometry"]
                dep = None
                depgeo = None

                try:
                    for j in dep_list.index:
                        depgeo = dep_list.loc[j, "geometry"]
                        if geo.intersects(depgeo):
                            dep = dep_list.loc[j, "insee_dep"]
                            break
                        else:
                            depgeo = None
                except Exception:
                    pass

                list_dep += [dep]
                list_dep_geo += [depgeo]

            gdf["insee_dep"] = list_dep
            gdf["insee_dep_geometry"] = list_dep_geo
    except Exception:
        pass

    return gdf



import re
import os
from functools import lru_cache

from pynsee.download._get_file_list_internal import _get_file_list_internal
from pynsee.utils.requests_session import PynseeAPISession

import logging

logger = logging.getLogger(__name__)


@lru_cache(maxsize=None)
def _get_dict_data_source():
    try:
        URL_DATA_LIST = os.environ["pynsee_file_list"]
    except KeyError:
        URL_MELODI = "https://minio.lab.sspcloud.fr/pierrelamarche/melodi/liste_donnees.json"
        URL_DOREMIFASOL = "https://raw.githubusercontent.com/InseeFrLab/DoReMIFaSol/master/data-raw/liste_donnees.json"  # inherited from previous pynsee version
        URL_DATA_LIST = [URL_MELODI] + [URL_DOREMIFASOL]

    with PynseeAPISession() as session:
        if isinstance(URL_DATA_LIST, list):
            try:
                jsonfile1 = session.get(URL_MELODI, verify=False).json()
                jsonfile2 = session.get(URL_DOREMIFASOL, verify=False).json()
                jsonfile = jsonfile1 + jsonfile2
            except Exception:
                logger.error("Error when reading sources catalog")
        else:
            try:
                jsonfile = [session.get(URL_DATA_LIST, verify=False).json()]
            except Exception:
                jsonfile = _get_file_list_internal()

    # HACK BECAUSE OF DUPLICATED ENTRIES -------------------------------

    potential_keys = [items["nom"] for items in jsonfile]
    list_duplicated_sources = list(
        set([x for x in potential_keys if potential_keys.count(x) > 1])
    )

    dict_data_source = {
        _create_key(item, list_duplicated_sources): item for item in jsonfile
    }

    return dict_data_source


def _create_key(item, duplicate_sources):
    """
    Transform JSON into Python dict

    Args:
        item: Item in JSON source
        duplicate_sources: List of duplicate sources in JSON

    Returns: Initial dict with modified keys to avoid duplicated entries

    """
    if item["nom"] not in duplicate_sources:
        return item["nom"]
    dateref = item["date_ref"]
    year = re.search(r"\d{4}", dateref).group(0)
    return f"{item['nom']}_{year}"



import zipfile
from pathlib import Path
from tqdm.utils import CallbackIOWrapper
from tqdm import tqdm
from shutil import copyfileobj
import os


def _unzip_pb(fzip, dest, desc="Extracting"):
    """
    Useful function to unzip with progress bar
    Args:
        fzip: Filename of the zipped file
        dest: Destination where data must be written
        desc: Argument inherited from zipfile.ZipFile

    Returns:
        zipfile.Zipfile(fzip).extractall(dest) with progress
    """

    dest = Path(dest).expanduser()
    Path(dest).mkdir(parents=True, exist_ok=True)

    with (
        zipfile.ZipFile(fzip) as zipf,
        tqdm(
            desc=desc,
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
            total=sum(getattr(i, "file_size", 0) for i in zipf.infolist()),
        ) as pbar,
    ):
        for i in zipf.infolist():
            if not getattr(i, "file_size", 0):  # directory
                zipf.extract(i, os.fspath(dest))
            else:
                with (
                    zipf.open(i) as tempfi,
                    open(os.fspath(dest / i.filename), "wb") as fileobj,
                ):
                    copyfileobj(
                        CallbackIOWrapper(pbar.update, tempfi), fileobj
                    )



import datetime
import functools
import logging
import os
import warnings

from typing import Optional, Type

import pandas as pd
import geopandas as gpd

from pynsee.utils._create_insee_folder import _create_insee_folder
from pynsee.utils._hash import _hash


logger = logging.getLogger(__name__)


@functools.lru_cache(maxsize=None)
def _warning_cached_data(file, mdate=None, day_lapse=None):
    strg_print = f"Previously saved data has been used:\n{file}\n"
    if (mdate is not None) and (day_lapse is not None):
        strg_print += f"Creation date: {mdate:%Y-%m-%d}"

        if day_lapse >= 2:
            strg_print += f", {day_lapse} days ago\n"
        if day_lapse == 1:
            strg_print += ", yesterday\n"
        if day_lapse == 0:
            strg_print += ", today\n"

    strg_print += "Set update=True to get the most up-to-date data"

    logger.info(strg_print)


def save_df(
    cls: Type[pd.DataFrame] = pd.DataFrame,
    day_lapse_max: Optional[int] = None,
):
    """
    Autosave object to reload from file unless it's empty.

    Note that if PYNSEE_SILENT_MODE is present in os.environ and set to "true",
    no logging.info are displayed to warn about the dataframe's storage nor the
    dataframe's retrieval from disk.

    Parameters
    ----------
    cls : Type[pd.DataFrame], optional
        The class used (should either be pd.DataFrame, SireneDataFrame,
        gpd.GeoDataFrame or GeoFrDataFrame). The default is pd.DataFrame.
    day_lapse_max : Optional[int], optional
        The maximum number of days to keep a file on disk before forcing a new
        download, whether or not update has been set to True. The default is
        None.

    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            data_folder = _create_insee_folder()

            string_file_arg = [
                str(kwargs[a])
                for a in kwargs.keys()
                if a not in ("update", "silent")
            ]

            string_file_arg += [func.__name__] + [str(a) for a in args]

            file_name = os.path.join(
                data_folder, _hash("".join(string_file_arg))
            )

            file_name += ".parquet"
            update = kwargs.get("update", False)

            silent_env = os.environ.get("PYNSEE_SILENT_MODE", "")
            silent = kwargs.get("silent", False) or (
                silent_env.lower() == "true"
            )

            if os.path.exists(file_name):
                file_date_last_modif = datetime.datetime.fromtimestamp(
                    os.path.getmtime(file_name)
                )

                insee_date_time_now = kwargs.get(
                    "insee_date_test", datetime.datetime.now()
                )

                day_lapse = (insee_date_time_now - file_date_last_modif).days

                if day_lapse_max is not None:
                    if day_lapse > day_lapse_max:
                        update = True

            if (not os.path.exists(file_name)) or update:
                df = func(*args, **kwargs)

                _save_dataframe(df, file_name, silent)
            else:
                try:
                    try:
                        mod = gpd if issubclass(cls, gpd.GeoDataFrame) else pd
                        df = mod.read_parquet(file_name)
                    except Exception:
                        df = pd.read_pickle(file_name)
                except Exception as e:
                    warnings.warn(str(e), stacklevel=2)

                    kwargs2 = kwargs
                    kwargs2["update"] = True

                    warnings.warn(
                        "!!! Unable to load data, recalling function "
                        "with `update` set to True !!!",
                        stacklevel=2,
                    )

                    df = func(*args, **kwargs2)

                    _save_dataframe(df, file_name, silent)
                else:
                    if not silent:
                        mdate = insee_date_time_now - datetime.timedelta(
                            days=day_lapse
                        )

                        _warning_cached_data(
                            file_name, mdate=mdate, day_lapse=day_lapse
                        )

            df.__class__ = cls

            return df

        return wrapper

    return decorator


def _save_dataframe(df: pd.DataFrame, file_name: str, silent: bool) -> None:
    """
    Triggers the storage of a dataframe on disk.

    Parameters
    ----------
    df : pd.DataFrame
        Dataset to store as a parquet file.
    file_name : str
        Dataset's filename.
    silent : bool
        If silent, will not log the storage.

    Returns
    -------
    None

    """
    try:
        if not df.empty:
            df.to_parquet(file_name)
    except Exception as e:
        warnings.warn(
            f"{e}\nError, file not saved:\n{file_name}\n{df}\n", stacklevel=2
        )
    else:
        if not silent:
            logger.info("Data saved:\n%s", file_name)



import warnings
from typing import Union

import numpy as np
from pandas import DataFrame
from geopandas import GeoDataFrame, GeoSeries


class GeoFrDataFrame(GeoDataFrame):
    """
    Class for handling GeoDataFrames built from IGN's geographical data.
    It inherits from :class:`~geopandas.GeoDataFrame`.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return GeoFrDataFrame

    def _constructor_from_mgr(
        self, mgr, axes
    ) -> Union[DataFrame, "GeoFrDataFrame"]:
        """Patch to avoid class change."""
        obj = super()._constructor_from_mgr(mgr, axes)

        if isinstance(obj, GeoDataFrame):
            obj.__class__ = GeoFrDataFrame

        return obj

    def __getitem__(self, key):
        """Patch to avoid class change."""
        result = super().__getitem__(key)

        if isinstance(result, GeoDataFrame):
            result.__class__ = GeoFrDataFrame

        return result

    def copy(self, deep: bool = True) -> "GeoFrDataFrame":
        """Copy the :class:`GeoFrDataFrame` object."""
        # Patch to avoid class change
        copied = super().copy(deep=deep)
        copied.__class__ = GeoFrDataFrame
        return copied

    def get_geom(self):
        """
        Return the combination of all geometries in the
        :class:`GeoFrDataFrame`.

        .. deprecated:: 0.2.0

            Use :attr:`~geopandas.GeoDataFrame.geometry` instead and call
            :func:`~geopandas.GeoSeries.union_all` on it.
            See also the `documentation of geopandas <https://geopandas.org/en/stable/docs/user_guide/data_structures.html#geodataframe>`_.

        Example
        -------

        >>> geo = gdf.geometry.union_all()
        """
        warnings.warn(
            "`get_geom` is deprecated, please use the `geometry` "
            "property of the `GeoFrDataFrame` instead. Then call "
            ":func:`~geopandas.GeoSeries.union_all` on it.",
            category=DeprecationWarning,
            stacklevel=2,
        )

        return self.geometry.union_all()

    def transform_overseas(self, *args, **kwargs) -> "GeoFrDataFrame":
        """
        Apply translation and zoom to oversea territories.

        See: :func:`pynsee.geodata.transform_overseas`.
        """
        from .translate_and_zoom import transform_overseas

        return transform_overseas(self, *args, **kwargs)

    def translate(self, *args, **kwargs) -> Union[GeoSeries, "GeoFrDataFrame"]:
        """
        This function is a deprecated alias of
        :meth:`~GeoFrDataFrame.transform_overseas`.
        It will try to guess whether you want to run `transform_overseas`
        or the real :meth:`GeoDataFrame.translate` depending on the
        arguments that were passed.
        If no arguments are passed, it will run `transform_overseas`.

        .. warning::

            Starting with ``pynsee >= 0.3.0``, this function will only
            run :meth:`GeoSeries.translate`. Do switch to
            :meth:`~GeoFrDataFrame.transform_overseas` if this is what
            you wanted to run.
        """
        run_super_translate = False

        if args and isinstance(args[0], (int, np.integer, float)):
            run_super_translate = True
        elif kwargs and ("xoff" in kwargs or "yoff" in kwargs):
            run_super_translate = True

        if run_super_translate:
            return super().translate(*args, **kwargs)

        warnings.warn(
            "`translate` will switch to the default `GeoDataFrame.translate` "
            "behavior in the next release. Please use `transform_overseas` "
            "instead.",
            category=DeprecationWarning,
            stacklevel=2,
        )

        return self.transform_overseas(*args, **kwargs)

    def zoom(self, *args, **kwargs) -> "GeoFrDataFrame":
        """
        Zoom for parisian departments.

        See: :func:`pynsee.geodata.zoom`.
        """
        from .translate_and_zoom import zoom

        return zoom(self, *args, **kwargs)



# -*- coding: utf-8 -*-
# Copyright : INSEE, 2021

from datetime import date
import io
import logging
import os
import re
import warnings
import zipfile

import pandas as pd

from ..utils.requests_session import PynseeAPISession
from ..utils.save_df import save_df


logger = logging.getLogger(__name__)


@save_df(day_lapse_max=90)
def _download_idbank_list(
    update: bool = False,
    silent: bool = False,
):
    """
    Retrieve the correspondance between idbank and its dimensions; the result
    is cached.

    Parameters
    ----------
    update : bool, optional
        If True, a new download will be performed whatever the current cache's
        status. The default is False.
    silent : bool, optional
        If silent, will not log the caching. The default is False.

    Returns
    -------
    data : pd.DataFrame
        correspondance between idbank and it's dimensions as DataFrame

    """
    data = _dwn_idbank_files()

    data.columns = ["nomflow", "idbank", "cleFlow", "list_var"]

    data = data.sort_values("nomflow").reset_index(drop=True)

    return data


def _dwn_idbank_files() -> pd.DataFrame:
    """
    Retrieve the correspondance between idbank and its dimensions of today's
    date.

    Note: if `pynsee_idbank_file` is set (through an os environment variable),
    this value will be used to perform the download instead of the potential
    URLs tested by the inner loop. This should only be used manually if the
    current algorithm fails to retrieve the actual URL.

    Returns
    -------
    data : pd.DataFrame
        correspondance between idbank and it's dimensions as DataFrame
    """

    todays_date = date.today()

    main_link_en = "https://www.insee.fr/en/statistiques/fichier/2868055/"
    main_link_fr = "https://www.insee.fr/fr/statistiques/fichier/2862759/"

    curr_year = todays_date.year
    last_year = curr_year - 1
    years = [str(curr_year), str(last_year)]

    months = [str(x) for x in range(12, 9, -1)] + [
        "0" + str(x) for x in range(9, 0, -1)
    ]

    today = todays_date.strftime("%Y%m")
    patt = "_correspondance_idbank_dimension"
    patterns = [
        y + x + patt for y in years for x in months if not y + x > today
    ]
    files_en = [main_link_en + f + ".zip" for f in patterns]
    files_fr = [main_link_fr + f + ".zip" for f in patterns]
    files = files_fr + files_en

    idbank_file_not_found = True
    if "pynsee_idbank_file" in os.environ:
        file_to_dwn = os.environ["pynsee_idbank_file"]
        with PynseeAPISession() as session:
            try:
                data = _dwn_idbank_file(
                    file_to_dwn=file_to_dwn, session=session
                )
                idbank_file_not_found = False
            except Exception:
                pass

    i = 0
    with PynseeAPISession() as session:

        while idbank_file_not_found and i < len(files):
            try:
                data = _dwn_idbank_file(file_to_dwn=files[i], session=session)
            except Exception:
                idbank_file_not_found = True
            else:
                idbank_file_not_found = False
                strg_print = f"Macrodata series update, file used:\n{files[i]}"
                logger.info(strg_print)
            i += 1

    return data


def _dwn_idbank_file(
    file_to_dwn: str, session: PynseeAPISession
) -> pd.DataFrame:
    """
    Download and load the correspondance between idbank and its dimensions,
    knowing it's actual URL.

    Parameters
    ----------
    file_to_dwn : str
        URL to download the file from.
    session : PynseeAPISession
        Current Session used to perform the http request (inherits from
        requests.Session)

    Returns
    -------
    data : pd.DataFrame
        correspondance between idbank and it's dimensions as DataFrame

    """
    separator = ";"

    proxies = {}
    for key in ["http", "https"]:
        try:
            proxies[key] = os.environ[f"{key}_proxy"]
        except KeyError:
            proxies[key] = ""

    with warnings.catch_warnings():
        results = session.get(file_to_dwn, proxies=proxies, verify=False)

    idbank_zip_file = io.BytesIO(results.content)

    with zipfile.ZipFile(idbank_zip_file) as zip_ref:
        file_to_read = [
            f for f in zip_ref.namelist() if not re.match(".*.zip$", f)
        ]
        if len(file_to_read) == 0:
            # nested zipfile
            new_zip_file = [
                f for f in zip_ref.namelist() if re.match(".*.zip$", f)
            ][0]

            with zip_ref.open(new_zip_file) as nested_file:
                read = nested_file.read()

            with zipfile.ZipFile(io.BytesIO(read)) as new_zip_ref:
                file_to_read = [
                    f
                    for f in new_zip_ref.namelist()
                    if not re.match(".*.zip$", f)
                ][0]
                with new_zip_ref.open(file_to_read) as f:
                    content = f.read()
        else:
            with zip_ref.open(file_to_read[0]) as f:
                content = f.read()

    file2load = io.BytesIO(content)
    data = pd.read_csv(file2load, dtype="str", sep=separator)

    return data



# -*- coding: utf-8 -*-

from pandas import DataFrame

from pynsee.utils.save_df import save_df

from ._get_full_list_wfs import _get_full_list_wfs


@save_df(day_lapse_max=90)
def get_geodata_list(update=False, silent=False) -> DataFrame:
    """Get a list of geographical limits of French administrative areas from IGN API

    Args:
        update (bool, optional): Trigger an update, otherwise locally saved data is used. Defaults to False.

        silent (bool, optional): Set to True, to disable messages printed in log info

    Examples:
        >>> from pynsee.geodata import get_geodata_list
        >>> # Get a list of geographical limits of French administrative areas from IGN API
        >>> geodata_list = get_geodata_list()
    """

    format = "WFS"
    topic = "administratif"
    version = "2.0.0"

    data_full_list = _get_full_list_wfs(topic=topic, version=version)

    if len(data_full_list) > 0:
        list_var = [
            "Name",
            "Identifier",
            "Title",
            "DefaultCRS",
            "SupportedCRS",
            "TileMatrixSet",
            "Abstract",
            "LegendURL",
            "Format",
        ]

        list_first_col = [
            col for col in data_full_list.columns if col in list_var
        ]
        list_other_col = [
            col for col in data_full_list.columns if col not in list_first_col
        ]

        data_list = data_full_list[list_first_col + list_other_col]
        data_list = data_list.drop_duplicates().reset_index(drop=True)

        if "Name" in data_list.columns:
            data_list.rename(columns={"Name": "Identifier"}, inplace=True)

        data_list["DataFormat"] = format
        data_list["ApiVersion"] = version

    data_all = data_list.reset_index(drop=True)

    # set column order
    first_col = [
        "Topic",
        "DataFormat",
        "ApiVersion",
        "Identifier",
        "Abstract",
        "Title",
        "ZoomRange",
    ]
    available_col = [col for col in first_col if col in data_all.columns]
    other_col = [col for col in data_all.columns if col not in available_col]

    data_all = data_all[available_col + other_col]

    return data_all



import difflib

from pynsee.geodata.get_geodata_list import get_geodata_list
from pynsee.utils.HiddenPrints import HiddenPrints

string = "ADMINEXPRESS-COG.LATEST:departement"


def _find_wfs_closest_match(string=string):
    with HiddenPrints():
        wfs = get_geodata_list()

    list_sugg = list(wfs.Identifier.unique())
    suggestions = difflib.get_close_matches(string, list_sugg)
    return suggestions[0]



# Copyright : INSEE, 2021

import pandas as pd
import math
from functools import lru_cache

from pynsee.utils.requests_session import PynseeAPISession
from pynsee.sirene._make_dataframe import _make_dataframe

import logging

logger = logging.getLogger(__name__)


@lru_cache(maxsize=None)
def _request_sirene(query, kind, number=1001):

    if kind == "siren":
        main_key = "unitesLegales"
    elif kind == "siret":
        main_key = "etablissements"
    else:
        raise ValueError("!!! kind should be among : siren siret !!!")

    INSEE_api_sirene_siren = "https://api.insee.fr/api-sirene/3.11"
    number_query_limit = 1000

    number_query = min(number_query_limit, number)

    n_query_total = math.ceil(number / number_query_limit)
    i_query = 1
    query_number = "{}/{}".format(i_query, n_query_total)

    main_query = INSEE_api_sirene_siren + "/" + kind + query

    link = main_query + "&nombre={}".format(number_query)

    if number > number_query_limit:
        link = link + "&curseur=*"

    with PynseeAPISession() as session:
        request = session.request_insee(
            api_url=link,
            file_format="application/json;charset=utf-8",
        )

    list_dataframe = []

    request_status = request.status_code

    if request_status == 200:
        data_request = request.json()

        data_request_1 = _make_dataframe(data_request, main_key, "1")

        if "siret" in data_request_1.columns:
            df_nrows = len(data_request_1.siret.unique())
        elif "siren" in data_request_1.columns:
            df_nrows = len(data_request_1.siren.unique())
        else:
            df_nrows = len(data_request_1.index.unique())

        list_dataframe.append(data_request_1)

        list_header_keys = list(data_request["header"].keys())

        if "curseur" in list_header_keys:
            cursor = data_request["header"]["curseur"]
            following_cursor = data_request["header"]["curseurSuivant"]

            #  & (i_query < query_limit)
            while (
                (following_cursor != cursor)
                & (request_status == 200)
                & (df_nrows < number)
            ):
                i_query += 1
                query_number = "{}/{}".format(i_query, n_query_total)

                new_query = (
                    main_query
                    + "&nombre={}".format(number_query_limit)
                    + "&curseur="
                    + following_cursor
                )

                with PynseeAPISession() as session:
                    request_new = session.request_insee(
                        api_url=new_query,
                        file_format="application/json;charset=utf-8",
                    )

                request_status = request_new.status_code

                if request_status == 200:
                    data_request_new = request_new.json()
                    cursor = data_request_new["header"]["curseur"]
                    following_cursor = data_request_new["header"][
                        "curseurSuivant"
                    ]

                    if len(data_request_new[main_key]) > 0:
                        df = _make_dataframe(
                            data_request_new, main_key, query_number
                        )

                        if "siret" in df.columns:
                            df_nrows += len(df.siret.unique())
                        elif "siren" in df.columns:
                            df_nrows += len(df.siren.unique())
                        else:
                            df_nrows += len(df.index.unique())

                        list_dataframe.append(df)
                    else:
                        logger.debug(
                            "{} - No more data found".format(query_number)
                        )

                    if cursor == following_cursor:
                        i_query += 1
                        query_number = "{}/{}".format(i_query, n_query_total)
                        logger.debug(
                            "{} - No more data found".format(query_number)
                        )

                    if df_nrows == number:
                        logger.warning(
                            "maximum reached, "
                            "increase value of number argument !"
                        )

        data_final = pd.concat(list_dataframe)

        if "siret" in data_final.columns:
            sirenCol = "siret"
        elif "siren" in data_final.columns:
            sirenCol = "siren"
        else:
            sirenCol = None

        if sirenCol is not None:
            if len(data_final[sirenCol].unique()) == number_query_limit:
                logger.warning(
                    "The query reached maximum item limit, "
                    "please change argument number to get more than "
                    f"{number_query_limit} different {sirenCol}"
                )

        return data_final
    else:
        logger.error(request.text)
        return None
